import os
# Configuration file for jupyter-server.

c = get_config()  # noqa: F821

#------------------------------------------------------------------------------
# Application(SingletonConfigurable) configuration
#------------------------------------------------------------------------------
## This is an application.

## The date format used by logging formatters for %(asctime)s
#  Default: '%Y-%m-%d %H:%M:%S'
# c.Application.log_datefmt = '%Y-%m-%d %H:%M:%S'

## The Logging format template
#  Default: '[%(name)s]%(highlevel)s %(message)s'
# c.Application.log_format = '[%(name)s]%(highlevel)s %(message)s'

## Set the log level by value or name.
#  Choices: any of [0, 10, 20, 30, 40, 50, 'DEBUG', 'INFO', 'WARN', 'ERROR', 'CRITICAL']
#  Default: 30
# c.Application.log_level = 30

## Instead of starting the Application, dump configuration to stdout
#  Default: False
# c.Application.show_config = False

## Instead of starting the Application, dump configuration to stdout (as JSON)
#  Default: False
# c.Application.show_config_json = False

#------------------------------------------------------------------------------
# JupyterApp(Application) configuration
#------------------------------------------------------------------------------
## Base class for Jupyter applications

## Answer yes to any prompts.
#  Default: False
# c.JupyterApp.answer_yes = False

## Full path of a config file.
#  Default: ''
# c.JupyterApp.config_file = ''

## Specify a config file to load.
#  Default: ''
# c.JupyterApp.config_file_name = ''

## Generate default config file.
#  Default: False
# c.JupyterApp.generate_config = False

## The date format used by logging formatters for %(asctime)s
#  See also: Application.log_datefmt
# c.JupyterApp.log_datefmt = '%Y-%m-%d %H:%M:%S'

## The Logging format template
#  See also: Application.log_format
# c.JupyterApp.log_format = '[%(name)s]%(highlevel)s %(message)s'

## Set the log level by value or name.
#  See also: Application.log_level
# c.JupyterApp.log_level = 30

## Instead of starting the Application, dump configuration to stdout
#  See also: Application.show_config
# c.JupyterApp.show_config = False

## Instead of starting the Application, dump configuration to stdout (as JSON)
#  See also: Application.show_config_json
# c.JupyterApp.show_config_json = False

#------------------------------------------------------------------------------
# ServerApp(JupyterApp) configuration
#------------------------------------------------------------------------------
## Set the Access-Control-Allow-Credentials: true header
#  Default: False
# c.ServerApp.allow_credentials = False

## Set the Access-Control-Allow-Origin header
#
#          Use '*' to allow any origin to access your server.
#
#          Takes precedence over allow_origin_pat.
#  Default: ''
# c.ServerApp.allow_origin = ''

## Use a regular expression for the Access-Control-Allow-Origin header
#
#          Requests from an origin matching the expression will get replies with:
#
#              Access-Control-Allow-Origin: origin
#
#          where `origin` is the origin of the request.
#
#          Ignored if allow_origin is set.
#  Default: ''
# c.ServerApp.allow_origin_pat = ''

## Allow password to be changed at login for the Jupyter server.
#
#                      While logging in with a token, the Jupyter server UI will give the opportunity to
#                      the user to enter a new password at the same time that will replace
#                      the token login mechanism.
#
#                      This can be set to false to prevent changing password from
#  the UI/API.
#  Default: True
# c.ServerApp.allow_password_change = True

## Allow requests where the Host header doesn't point to a local server
#
#         By default, requests get a 403 forbidden response if the 'Host' header
#         shows that the browser thinks it's on a non-local domain.
#         Setting this option to True disables this check.
#
#         This protects against 'DNS rebinding' attacks, where a remote web server
#         serves you a page and then changes its DNS to send later requests to a
#         local IP, bypassing same-origin checks.
#
#         Local IP addresses (such as 127.0.0.1 and ::1) are allowed as local,
#         along with hostnames configured in local_hostnames.
#  Default: False
# c.ServerApp.allow_remote_access = False

## Whether to allow the user to run the server as root.
#  Default: False
# c.ServerApp.allow_root = False
# Whether to allow the user to run the notebook as root.
c.ServerApp.allow_root = True if os.environ.get('JY_ALLOW_ROOT') else False


## Answer yes to any prompts.
#  See also: JupyterApp.answer_yes
# c.ServerApp.answer_yes = False

## "
#          Require authentication to access prometheus metrics.
#  Default: True
# c.ServerApp.authenticate_prometheus = True

## Reload the webapp when changes are made to any Python src files.
#  Default: False
# c.ServerApp.autoreload = False

## The base URL for the Jupyter server.
#
#                         Leading and trailing slashes can be omitted,
#                         and will automatically be added.
#  Default: '/'
# c.ServerApp.base_url = '/'

## Specify what command to use to invoke a web
#                        browser when starting the server. If not specified, the
#                        default browser will be determined by the `webbrowser`
#                        standard library module, which allows setting of the
#                        BROWSER environment variable to override it.
#  Default: ''
# c.ServerApp.browser = ''

## The full path to an SSL/TLS certificate file.
#  Default: ''
# c.ServerApp.certfile = ''

## The full path to a certificate authority certificate for SSL/TLS client
#  authentication.
#  Default: ''
# c.ServerApp.client_ca = ''

## Full path of a config file.
#  See also: JupyterApp.config_file
# c.ServerApp.config_file = ''

## Specify a config file to load.
#  See also: JupyterApp.config_file_name
# c.ServerApp.config_file_name = ''

## The config manager class to use
#  Default: 'jupyter_server.services.config.manager.ConfigManager'
# c.ServerApp.config_manager_class = 'jupyter_server.services.config.manager.ConfigManager'

## The content manager class to use.
#  Default: 'jupyter_server.services.contents.largefilemanager.LargeFileManager'
# c.ServerApp.contents_manager_class = 'jupyter_server.services.contents.largefilemanager.LargeFileManager'
default_contents_manager = 'omegaml.notebook.omegacontentsmgr.OmegaStoreContentsManager'
contents_manager = os.environ.get('JY_CONTENTS_MANAGER') or default_contents_manager
c.ServerApp.contents_manager_class = contents_manager

## Extra keyword arguments to pass to `set_secure_cookie`. See tornado's
#  set_secure_cookie docs for details.
#  Default: {}
# c.ServerApp.cookie_options = {}

## The random bytes used to secure cookies.
#          By default this is a new random number every time you start the server.
#          Set it to a value in a config file to enable logins to persist across server sessions.
#
#          Note: Cookie secrets should be kept private, do not share config files with
#          cookie_secret stored in plaintext (you can read the value from a file).
#  Default: b''
# c.ServerApp.cookie_secret = b''

## The file where the cookie secret is stored.
#  Default: ''
# c.ServerApp.cookie_secret_file = ''

## Override URL shown to users.
#
#          Replace actual URL, including protocol, address, port and base URL,
#          with the given value when displaying URL to the users. Do not change
#          the actual connection URL. If authentication token is enabled, the
#          token is added to the custom URL automatically.
#
#          This option is intended to be used when the URL to display to the user
#          cannot be determined reliably by the Jupyter server (proxified
#          or containerized setups for example).
#  Default: ''
# c.ServerApp.custom_display_url = ''

## The default URL to redirect to from `/`
#  Default: '/'
# c.ServerApp.default_url = '/'

## Disable cross-site-request-forgery protection
#
#          Jupyter notebook 4.3.1 introduces protection from cross-site request forgeries,
#          requiring API requests to either:
#
#          - originate from pages served by this server (validated with XSRF cookie and token), or
#          - authenticate with a token
#
#          Some anonymous compute resources still desire the ability to run code,
#          completely without authentication.
#          These services can disable all authentication and security checks,
#          with the full knowledge of what that implies.
#  Default: False
# c.ServerApp.disable_check_xsrf = False

## handlers that should be loaded at higher priority than the default services
#  Default: []
# c.ServerApp.extra_services = []

## Extra paths to search for serving static files.
#
#          This allows adding javascript/css to be available from the Jupyter server machine,
#          or overriding individual files in the IPython
#  Default: []
# c.ServerApp.extra_static_paths = []

## Extra paths to search for serving jinja templates.
#
#          Can be used to override templates from jupyter_server.templates.
#  Default: []
# c.ServerApp.extra_template_paths = []

## Open the named file when the application is launched.
#  Default: ''
# c.ServerApp.file_to_run = ''

## The URL prefix where files are opened directly.
#  Default: 'notebooks'
# c.ServerApp.file_url_prefix = 'notebooks'

## Generate default config file.
#  See also: JupyterApp.generate_config
# c.ServerApp.generate_config = False

## Extra keyword arguments to pass to `get_secure_cookie`. See tornado's
#  get_secure_cookie docs for details.
#  Default: {}
# c.ServerApp.get_secure_cookie_kwargs = {}

## (bytes/sec)
#          Maximum rate at which stream output can be sent on iopub before they are
#          limited.
#  Default: 1000000
# c.ServerApp.iopub_data_rate_limit = 1000000

## (msgs/sec)
#          Maximum rate at which messages can be sent on iopub before they are
#          limited.
#  Default: 1000
# c.ServerApp.iopub_msg_rate_limit = 1000

## The IP address the Jupyter server will listen on.
#  Default: 'localhost'
# c.ServerApp.ip = 'localhost'

## Supply extra arguments that will be passed to Jinja environment.
#  Default: {}
# c.ServerApp.jinja_environment_options = {}

## Extra variables to supply to jinja templates when rendering.
#  Default: {}
# c.ServerApp.jinja_template_vars = {}

## Dict of Python modules to load as Jupyter server extensions.Entry values can
#  be used to enable and disable the loading ofthe extensions. The extensions
#  will be loaded in alphabetical order.
#  Default: {}
# c.ServerApp.jpserver_extensions = {}

## The kernel manager class to use.
#  Default: 'jupyter_server.services.kernels.kernelmanager.AsyncMappingKernelManager'
# c.ServerApp.kernel_manager_class = 'jupyter_server.services.kernels.kernelmanager.AsyncMappingKernelManager'

## The kernel spec manager class to use. Should be a subclass of
#  `jupyter_client.kernelspec.KernelSpecManager`.
#
#  The Api of KernelSpecManager is provisional and might change without warning
#  between this version of Jupyter and the next stable one.
#  Default: 'jupyter_client.kernelspec.KernelSpecManager'
# c.ServerApp.kernel_spec_manager_class = 'jupyter_client.kernelspec.KernelSpecManager'

## The full path to a private key file for usage with SSL/TLS.
#  Default: ''
# c.ServerApp.keyfile = ''

## Hostnames to allow as local when allow_remote_access is False.
#
#         Local IP addresses (such as 127.0.0.1 and ::1) are automatically accepted
#         as local as well.
#  Default: ['localhost']
# c.ServerApp.local_hostnames = ['localhost']

## The date format used by logging formatters for %(asctime)s
#  See also: Application.log_datefmt
# c.ServerApp.log_datefmt = '%Y-%m-%d %H:%M:%S'

## The Logging format template
#  See also: Application.log_format
# c.ServerApp.log_format = '[%(name)s]%(highlevel)s %(message)s'

## Set the log level by value or name.
#  See also: Application.log_level
# c.ServerApp.log_level = 30

## The login handler class to use.
#  Default: 'jupyter_server.auth.login.LoginHandler'
# c.ServerApp.login_handler_class = 'jupyter_server.auth.login.LoginHandler'

## The logout handler class to use.
#  Default: 'jupyter_server.auth.logout.LogoutHandler'
# c.ServerApp.logout_handler_class = 'jupyter_server.auth.logout.LogoutHandler'

## Sets the maximum allowed size of the client request body, specified in the
#  Content-Length request header field. If the size in a request exceeds the
#  configured value, a malformed HTTP message is returned to the client.
#
#  Note: max_body_size is applied even in streaming mode.
#  Default: 536870912
# c.ServerApp.max_body_size = 536870912

## Gets or sets the maximum amount of memory, in bytes, that is allocated for use
#  by the buffer manager.
#  Default: 536870912
# c.ServerApp.max_buffer_size = 536870912

## Gets or sets a lower bound on the open file handles process resource limit.
#  This may need to be increased if you run into an OSError: [Errno 24] Too many
#  open files. This is not applicable when running on Windows.
#  Default: 0
# c.ServerApp.min_open_files_limit = 0

## DEPRECATED, use root_dir.
#  Default: ''
# c.ServerApp.notebook_dir = ''

## Whether to open in a browser after starting.
#                          The specific browser used is platform dependent and
#                          determined by the python standard library `webbrowser`
#                          module, unless it is overridden using the --browser
#                          (ServerApp.browser) configuration option.
#  Default: False
# c.ServerApp.open_browser = False

## Hashed password to use for web authentication.
#
#                        To generate, type in a python/IPython shell:
#
#                          from jupyter_server.auth import passwd; passwd()
#
#                        The string should be of the form type:salt:hashed-
#  password.
#  Default: ''
# c.ServerApp.password = ''
if 'JUPYTER_PASSWORD' in os.environ:
    # unless it is actually a value in environ we must not set the attribute, as None is not a valid value
    # if the value is not set a token is generated which is what we want in this case
    c.ServerApp.password = os.environ.get('JUPYTER_PASSWORD')

## Forces users to use a password for the Jupyter server.
#                        This is useful in a multi user environment, for instance when
#                        everybody in the LAN can access each other's machine through ssh.
#
#                        In such a case, serving on localhost is not secure since
#                        any user can connect to the Jupyter server via ssh.
#  Default: False
# c.ServerApp.password_required = False
c.ServerApp.password_required = 'JUPYTER_PASSWORD' in os.environ

## The port the server will listen on (env: JUPYTER_PORT).
#  Default: 0
c.ServerApp.port = int(os.environ.get('JUPYTER_PORT', 8888))

## The number of additional ports to try if the specified port is not available
#  (env: JUPYTER_PORT_RETRIES).
#  Default: 50
# c.ServerApp.port_retries = 50

## Preferred starting directory to use for notebooks and kernels.
#  Default: ''
# c.ServerApp.preferred_dir = ''

## DISABLED: use %pylab or %matplotlib in the notebook to enable matplotlib.
#  Default: 'disabled'
# c.ServerApp.pylab = 'disabled'

## If True, display controls to shut down the Jupyter server, such as menu items
#  or buttons.
#  Default: True
# c.ServerApp.quit_button = True

## (sec) Time window used to
#          check the message and data rate limits.
#  Default: 3
# c.ServerApp.rate_limit_window = 3

## Reraise exceptions encountered loading server extensions?
#  Default: False
# c.ServerApp.reraise_server_extension_failures = False

## The directory to use for notebooks and kernels.
#  Default: ''
# c.ServerApp.root_dir = ''
c.ServerApp.root_dir = os.getcwd()

## The session manager class to use.
#  Default: 'jupyter_server.services.sessions.sessionmanager.SessionManager'
# c.ServerApp.session_manager_class = 'jupyter_server.services.sessions.sessionmanager.SessionManager'

## Instead of starting the Application, dump configuration to stdout
#  See also: Application.show_config
# c.ServerApp.show_config = False

## Instead of starting the Application, dump configuration to stdout (as JSON)
#  See also: Application.show_config_json
# c.ServerApp.show_config_json = False

## Shut down the server after N seconds with no kernels or terminals running and
#  no activity. This can be used together with culling idle kernels
#  (MappingKernelManager.cull_idle_timeout) to shutdown the Jupyter server when
#  it's not in use. This is not precisely timed: it may shut down up to a minute
#  later. 0 (the default) disables this automatic shutdown.
#  Default: 0
# c.ServerApp.shutdown_no_activity_timeout = 0

## The UNIX socket the Jupyter server will listen on.
#  Default: ''
# c.ServerApp.sock = ''

## The permissions mode for UNIX socket creation (default: 0600).
#  Default: '0600'
# c.ServerApp.sock_mode = '0600'

## Supply SSL options for the tornado HTTPServer.
#              See the tornado docs for details.
#  Default: {}
# c.ServerApp.ssl_options = {}

## Supply overrides for terminado. Currently only supports "shell_command".
#  Default: {}
# c.ServerApp.terminado_settings = {}

## Set to False to disable terminals.
#
#           This does *not* make the server more secure by itself.
#           Anything the user can in a terminal, they can also do in a notebook.
#
#           Terminals may also be automatically disabled if the terminado package
#           is not available.
#  Default: True
# c.ServerApp.terminals_enabled = True

## Token used for authenticating first-time connections to the server.
#
#          The token can be read from the file referenced by JUPYTER_TOKEN_FILE or set directly
#          with the JUPYTER_TOKEN environment variable.
#
#          When no password is enabled,
#          the default is to generate a new, random token.
#
#          Setting to an empty string disables authentication altogether, which
#  is NOT RECOMMENDED.
#  Default: '<generated>'
# c.ServerApp.token = '<generated>'

## Supply overrides for the tornado.web.Application that the Jupyter server uses.
#  Default: {}
# c.ServerApp.tornado_settings = {}

## Whether to trust or not X-Scheme/X-Forwarded-Proto and X-Real-Ip/X-Forwarded-
#  For headerssent by the upstream reverse proxy. Necessary if the proxy handles
#  SSL
#  Default: False
# c.ServerApp.trust_xheaders = False

## Disable launching browser by redirect file
#       For versions of notebook > 5.7.2, a security feature measure was added that
#       prevented the authentication token used to launch the browser from being visible.
#       This feature makes it difficult for other users on a multi-user system from
#       running code in your Jupyter session as you.
#       However, some environments (like Windows Subsystem for Linux (WSL) and Chromebooks),
#       launching a browser using a redirect file can lead the browser failing to load.
#       This is because of the difference in file structures/paths between the runtime and
#       the browser.
#
#       Disabling this setting to False will disable this behavior, allowing the browser
#       to launch by using a URL and visible token (as before).
#  Default: True
# c.ServerApp.use_redirect_file = True

## Specify where to open the server on startup. This is the
#          `new` argument passed to the standard library method `webbrowser.open`.
#          The behaviour is not guaranteed, but depends on browser support. Valid
#          values are:
#
#           - 2 opens a new tab,
#           - 1 opens a new window,
#           - 0 opens in an existing window.
#
#          See the `webbrowser.open` documentation for details.
#  Default: 2
# c.ServerApp.webbrowser_open_new = 2

## Set the tornado compression options for websocket connections.
#
#  This value will be returned from
#  :meth:`WebSocketHandler.get_compression_options`. None (default) will disable
#  compression. A dict (even an empty one) will enable compression.
#
#  See the tornado docs for WebSocketHandler.get_compression_options for details.
#  Default: None
# c.ServerApp.websocket_compression_options = None

## The base URL for websockets,
#          if it differs from the HTTP server (hint: it almost certainly doesn't).
#
#          Should be in the form of an HTTP origin: ws[s]://hostname[:port]
#  Default: ''
# c.ServerApp.websocket_url = ''

#------------------------------------------------------------------------------
# ConnectionFileMixin(LoggingConfigurable) configuration
#------------------------------------------------------------------------------
## Mixin for configurable classes that work with connection files

## JSON file in which to store connection info [default: kernel-<pid>.json]
#
#      This file will contain the IP, ports, and authentication key needed to connect
#      clients to this kernel. By default, this file will be created in the security dir
#      of the current profile, but can be specified by absolute path.
#  Default: ''
# c.ConnectionFileMixin.connection_file = ''

## set the control (ROUTER) port [default: random]
#  Default: 0
# c.ConnectionFileMixin.control_port = 0

## set the heartbeat port [default: random]
#  Default: 0
# c.ConnectionFileMixin.hb_port = 0

## set the iopub (PUB) port [default: random]
#  Default: 0
# c.ConnectionFileMixin.iopub_port = 0

## Set the kernel's IP address [default localhost].
#          If the IP address is something other than localhost, then
#          Consoles on other machines will be able to connect
#          to the Kernel, so be careful!
#  Default: ''
# c.ConnectionFileMixin.ip = ''

## set the shell (ROUTER) port [default: random]
#  Default: 0
# c.ConnectionFileMixin.shell_port = 0

## set the stdin (ROUTER) port [default: random]
#  Default: 0
# c.ConnectionFileMixin.stdin_port = 0

#  Choices: any of ['tcp', 'ipc'] (case-insensitive)
#  Default: 'tcp'
# c.ConnectionFileMixin.transport = 'tcp'

#------------------------------------------------------------------------------
# KernelManager(ConnectionFileMixin) configuration
#------------------------------------------------------------------------------
## Manages a single kernel in a subprocess on this host.
#
#      This version starts kernels with Popen.

## Should we autorestart the kernel if it dies.
#  Default: True
# c.KernelManager.autorestart = True

## JSON file in which to store connection info [default: kernel-<pid>.json]
#  See also: ConnectionFileMixin.connection_file
# c.KernelManager.connection_file = ''

## set the control (ROUTER) port [default: random]
#  See also: ConnectionFileMixin.control_port
# c.KernelManager.control_port = 0

## set the heartbeat port [default: random]
#  See also: ConnectionFileMixin.hb_port
# c.KernelManager.hb_port = 0

## set the iopub (PUB) port [default: random]
#  See also: ConnectionFileMixin.iopub_port
# c.KernelManager.iopub_port = 0

## Set the kernel's IP address [default localhost].
#  See also: ConnectionFileMixin.ip
# c.KernelManager.ip = ''

## set the shell (ROUTER) port [default: random]
#  See also: ConnectionFileMixin.shell_port
# c.KernelManager.shell_port = 0

## Time to wait for a kernel to terminate before killing it, in seconds. When a
#  shutdown request is initiated, the kernel will be immediately sent an
#  interrupt (SIGINT), followedby a shutdown_request message, after 1/2 of
#  `shutdown_wait_time`it will be sent a terminate (SIGTERM) request, and finally
#  at the end of `shutdown_wait_time` will be killed (SIGKILL). terminate and
#  kill may be equivalent on windows.  Note that this value can beoverridden by
#  the in-use kernel provisioner since shutdown times mayvary by provisioned
#  environment.
#  Default: 5.0
# c.KernelManager.shutdown_wait_time = 5.0

## set the stdin (ROUTER) port [default: random]
#  See also: ConnectionFileMixin.stdin_port
# c.KernelManager.stdin_port = 0

#  See also: ConnectionFileMixin.transport
# c.KernelManager.transport = 'tcp'

#------------------------------------------------------------------------------
# Session(Configurable) configuration
#------------------------------------------------------------------------------
## Object for handling serialization and sending of messages.
#
#      The Session object handles building messages and sending them
#      with ZMQ sockets or ZMQStream objects.  Objects can communicate with each
#      other over the network via Session objects, and only need to work with the
#      dict-based IPython message spec. The Session will handle
#      serialization/deserialization, security, and metadata.
#
#      Sessions support configurable serialization via packer/unpacker traits,
#      and signing with HMAC digests via the key/keyfile traits.
#
#      Parameters
#      ----------
#
#      debug : bool
#          whether to trigger extra debugging statements
#      packer/unpacker : str : 'json', 'pickle' or import_string
#          importstrings for methods to serialize message parts.  If just
#          'json' or 'pickle', predefined JSON and pickle packers will be used.
#          Otherwise, the entire importstring must be used.
#
#          The functions must accept at least valid JSON input, and output
#  *bytes*.
#
#          For example, to use msgpack:
#          packer = 'msgpack.packb', unpacker='msgpack.unpackb'
#      pack/unpack : callables
#          You can also set the pack/unpack callables for serialization directly.
#      session : bytes
#          the ID of this Session object.  The default is to generate a new UUID.
#      username : unicode
#          username added to message headers.  The default is to ask the OS.
#      key : bytes
#          The key used to initialize an HMAC signature.  If unset, messages
#          will not be signed or checked.
#      keyfile : filepath
#          The file containing a key.  If this is set, `key` will be initialized
#          to the contents of the file.

## Threshold (in bytes) beyond which an object's buffer should be extracted to
#  avoid pickling.
#  Default: 1024
# c.Session.buffer_threshold = 1024

## Whether to check PID to protect against calls after fork.
#
#          This check can be disabled if fork-safety is handled elsewhere.
#  Default: True
# c.Session.check_pid = True

## Threshold (in bytes) beyond which a buffer should be sent without copying.
#  Default: 65536
# c.Session.copy_threshold = 65536

## Debug output in the Session
#  Default: False
# c.Session.debug = False

## The maximum number of digests to remember.
#
#          The digest history will be culled when it exceeds this value.
#  Default: 65536
# c.Session.digest_history_size = 65536

## The maximum number of items for a container to be introspected for custom serialization.
#          Containers larger than this are pickled outright.
#  Default: 64
# c.Session.item_threshold = 64

## execution key, for signing messages.
#  Default: b''
# c.Session.key = b''

## path to file containing execution key.
#  Default: ''
# c.Session.keyfile = ''

## Metadata dictionary, which serves as the default top-level metadata dict for
#  each message.
#  Default: {}
# c.Session.metadata = {}

## The name of the packer for serializing messages.
#              Should be one of 'json', 'pickle', or an import name
#              for a custom callable serializer.
#  Default: 'json'
# c.Session.packer = 'json'

## The UUID identifying this session.
#  Default: ''
# c.Session.session = ''

## The digest scheme used to construct the message signatures.
#          Must have the form 'hmac-HASH'.
#  Default: 'hmac-sha256'
# c.Session.signature_scheme = 'hmac-sha256'

## The name of the unpacker for unserializing messages.
#          Only used with custom functions for `packer`.
#  Default: 'json'
# c.Session.unpacker = 'json'

## Username for the Session. Default is your system username.
#  Default: 'patrick'
# c.Session.username = 'patrick'

#------------------------------------------------------------------------------
# MultiKernelManager(LoggingConfigurable) configuration
#------------------------------------------------------------------------------
## A class for managing multiple kernels.

## The name of the default kernel to start
#  Default: 'python3'
# c.MultiKernelManager.default_kernel_name = 'python3'

## The kernel manager class.  This is configurable to allow
#          subclassing of the KernelManager for customized behavior.
#  Default: 'jupyter_client.ioloop.IOLoopKernelManager'
# c.MultiKernelManager.kernel_manager_class = 'jupyter_client.ioloop.IOLoopKernelManager'

## Share a single zmq.Context to talk to all my kernels
#  Default: True
# c.MultiKernelManager.shared_context = True

#------------------------------------------------------------------------------
# MappingKernelManager(MultiKernelManager) configuration
#------------------------------------------------------------------------------
## A KernelManager that handles
#      - File mapping
#      - HTTP error handling
#      - Kernel message filtering

## Whether to send tracebacks to clients on exceptions.
#  Default: True
# c.MappingKernelManager.allow_tracebacks = True

## White list of allowed kernel message types.
#          When the list is empty, all message types are allowed.
#  Default: []
# c.MappingKernelManager.allowed_message_types = []

## Whether messages from kernels whose frontends have disconnected should be
#  buffered in-memory.
#
#          When True (default), messages are buffered and replayed on reconnect,
#          avoiding lost messages due to interrupted connectivity.
#
#          Disable if long-running kernels will produce too much output while
#          no frontends are connected.
#  Default: True
# c.MappingKernelManager.buffer_offline_messages = True

## Whether to consider culling kernels which are busy.
#          Only effective if cull_idle_timeout > 0.
#  Default: False
# c.MappingKernelManager.cull_busy = False

## Whether to consider culling kernels which have one or more connections.
#          Only effective if cull_idle_timeout > 0.
#  Default: False
# c.MappingKernelManager.cull_connected = False

## Timeout (in seconds) after which a kernel is considered idle and ready to be culled.
#          Values of 0 or lower disable culling. Very short timeouts may result in kernels being culled
#          for users with poor network connections.
#  Default: 0
# c.MappingKernelManager.cull_idle_timeout = 0

## The interval (in seconds) on which to check for idle kernels exceeding the
#  cull timeout value.
#  Default: 300
# c.MappingKernelManager.cull_interval = 300

## The name of the default kernel to start
#  See also: MultiKernelManager.default_kernel_name
# c.MappingKernelManager.default_kernel_name = 'python3'

## Timeout for giving up on a kernel (in seconds).
#
#          On starting and restarting kernels, we check whether the
#          kernel is running and responsive by sending kernel_info_requests.
#          This sets the timeout in seconds for how long the kernel can take
#          before being presumed dead.
#          This affects the MappingKernelManager (which handles kernel restarts)
#          and the ZMQChannelsHandler (which handles the startup).
#  Default: 60
# c.MappingKernelManager.kernel_info_timeout = 60

## The kernel manager class.  This is configurable to allow
#  See also: MultiKernelManager.kernel_manager_class
# c.MappingKernelManager.kernel_manager_class = 'jupyter_client.ioloop.IOLoopKernelManager'

#  Default: ''
# c.MappingKernelManager.root_dir = ''

## Share a single zmq.Context to talk to all my kernels
#  See also: MultiKernelManager.shared_context
# c.MappingKernelManager.shared_context = True

## Message to print when allow_tracebacks is False, and an exception occurs
#  Default: 'An exception occurred at runtime, which is not shown due to security reasons.'
# c.MappingKernelManager.traceback_replacement_message = 'An exception occurred at runtime, which is not shown due to security reasons.'

#------------------------------------------------------------------------------
# KernelSpecManager(LoggingConfigurable) configuration
#------------------------------------------------------------------------------
## List of allowed kernel names.
#
#          By default, all installed kernels are allowed.
#  Default: set()
# c.KernelSpecManager.allowed_kernelspecs = set()

## If there is no Python kernelspec registered and the IPython
#          kernel is available, ensure it is added to the spec list.
#  Default: True
# c.KernelSpecManager.ensure_native_kernel = True

## The kernel spec class.  This is configurable to allow
#          subclassing of the KernelSpecManager for customized behavior.
#  Default: 'jupyter_client.kernelspec.KernelSpec'
# c.KernelSpecManager.kernel_spec_class = 'jupyter_client.kernelspec.KernelSpec'

## Deprecated, use `KernelSpecManager.allowed_kernelspecs`
#  Default: set()
# c.KernelSpecManager.whitelist = set()

#------------------------------------------------------------------------------
# AsyncMultiKernelManager(MultiKernelManager) configuration
#------------------------------------------------------------------------------
## The name of the default kernel to start
#  See also: MultiKernelManager.default_kernel_name
# c.AsyncMultiKernelManager.default_kernel_name = 'python3'

## The kernel manager class.  This is configurable to allow
#          subclassing of the AsyncKernelManager for customized behavior.
#  Default: 'jupyter_client.ioloop.AsyncIOLoopKernelManager'
# c.AsyncMultiKernelManager.kernel_manager_class = 'jupyter_client.ioloop.AsyncIOLoopKernelManager'

## Share a single zmq.Context to talk to all my kernels
#  See also: MultiKernelManager.shared_context
# c.AsyncMultiKernelManager.shared_context = True

#------------------------------------------------------------------------------
# AsyncMappingKernelManager(MappingKernelManager, AsyncMultiKernelManager) configuration
#------------------------------------------------------------------------------
## Whether to send tracebacks to clients on exceptions.
#  See also: MappingKernelManager.allow_tracebacks
# c.AsyncMappingKernelManager.allow_tracebacks = True

## White list of allowed kernel message types.
#  See also: MappingKernelManager.allowed_message_types
# c.AsyncMappingKernelManager.allowed_message_types = []

## Whether messages from kernels whose frontends have disconnected should be
#  buffered in-memory.
#  See also: MappingKernelManager.buffer_offline_messages
# c.AsyncMappingKernelManager.buffer_offline_messages = True

## Whether to consider culling kernels which are busy.
#  See also: MappingKernelManager.cull_busy
# c.AsyncMappingKernelManager.cull_busy = False

## Whether to consider culling kernels which have one or more connections.
#  See also: MappingKernelManager.cull_connected
# c.AsyncMappingKernelManager.cull_connected = False

## Timeout (in seconds) after which a kernel is considered idle and ready to be
#  culled.
#  See also: MappingKernelManager.cull_idle_timeout
# c.AsyncMappingKernelManager.cull_idle_timeout = 0

## The interval (in seconds) on which to check for idle kernels exceeding the
#  cull timeout value.
#  See also: MappingKernelManager.cull_interval
# c.AsyncMappingKernelManager.cull_interval = 300

## The name of the default kernel to start
#  See also: MultiKernelManager.default_kernel_name
# c.AsyncMappingKernelManager.default_kernel_name = 'python3'

## Timeout for giving up on a kernel (in seconds).
#  See also: MappingKernelManager.kernel_info_timeout
# c.AsyncMappingKernelManager.kernel_info_timeout = 60

## The kernel manager class.  This is configurable to allow
#  See also: AsyncMultiKernelManager.kernel_manager_class
# c.AsyncMappingKernelManager.kernel_manager_class = 'jupyter_client.ioloop.AsyncIOLoopKernelManager'

#  See also: MappingKernelManager.root_dir
# c.AsyncMappingKernelManager.root_dir = ''

## Share a single zmq.Context to talk to all my kernels
#  See also: MultiKernelManager.shared_context
# c.AsyncMappingKernelManager.shared_context = True

## Message to print when allow_tracebacks is False, and an exception occurs
#  See also: MappingKernelManager.traceback_replacement_message
# c.AsyncMappingKernelManager.traceback_replacement_message = 'An exception occurred at runtime, which is not shown due to security reasons.'

#------------------------------------------------------------------------------
# ContentsManager(LoggingConfigurable) configuration
#------------------------------------------------------------------------------
## Base class for serving files and directories.
#
#      This serves any text or binary file,
#      as well as directories,
#      with special handling for JSON notebook documents.
#
#      Most APIs take a path argument,
#      which is always an API-style unicode path,
#      and always refers to a directory.
#
#      - unicode, not url-escaped
#      - '/'-separated
#      - leading and trailing '/' will be stripped
#      - if unspecified, path defaults to '',
#        indicating the root path.

## Allow access to hidden files
#  Default: False
# c.ContentsManager.allow_hidden = False

#  Default: None
# c.ContentsManager.checkpoints = None

#  Default: 'jupyter_server.services.contents.checkpoints.Checkpoints'
# c.ContentsManager.checkpoints_class = 'jupyter_server.services.contents.checkpoints.Checkpoints'

#  Default: {}
# c.ContentsManager.checkpoints_kwargs = {}

## handler class to use when serving raw file requests.
#
#          Default is a fallback that talks to the ContentsManager API,
#          which may be inefficient, especially for large files.
#
#          Local files-based ContentsManagers can use a StaticFileHandler subclass,
#          which will be much more efficient.
#
#          Access to these files should be Authenticated.
#  Default: 'jupyter_server.files.handlers.FilesHandler'
# c.ContentsManager.files_handler_class = 'jupyter_server.files.handlers.FilesHandler'

## Extra parameters to pass to files_handler_class.
#
#          For example, StaticFileHandlers generally expect a `path` argument
#          specifying the root directory from which to serve files.
#  Default: {}
# c.ContentsManager.files_handler_params = {}

## Glob patterns to hide in file and directory listings.
#  Default: ['__pycache__', '*.pyc', '*.pyo', '.DS_Store', '*.so', '*.dylib', '*~']
# c.ContentsManager.hide_globs = ['__pycache__', '*.pyc', '*.pyo', '.DS_Store', '*.so', '*.dylib', '*~']

## Python callable or importstring thereof
#
#          To be called on a contents model prior to save.
#
#          This can be used to process the structure,
#          such as removing notebook outputs or other side effects that
#          should not be saved.
#
#          It will be called as (all arguments passed by keyword)::
#
#              hook(path=path, model=model, contents_manager=self)
#
#          - model: the model to be saved. Includes file contents.
#            Modifying this dict will affect the file that is stored.
#          - path: the API path of the save destination
#          - contents_manager: this ContentsManager instance
#  Default: None
# c.ContentsManager.pre_save_hook = None

#  Default: '/'
# c.ContentsManager.root_dir = '/'

## The base name used when creating untitled directories.
#  Default: 'Untitled Folder'
# c.ContentsManager.untitled_directory = 'Untitled Folder'

## The base name used when creating untitled files.
#  Default: 'untitled'
# c.ContentsManager.untitled_file = 'untitled'

## The base name used when creating untitled notebooks.
#  Default: 'Untitled'
# c.ContentsManager.untitled_notebook = 'Untitled'

#------------------------------------------------------------------------------
# FileManagerMixin(Configurable) configuration
#------------------------------------------------------------------------------
## Mixin for ContentsAPI classes that interact with the filesystem.
#
#  Provides facilities for reading, writing, and copying files.
#
#  Shared by FileContentsManager and FileCheckpoints.
#
#  Note ---- Classes using this mixin must provide the following attributes:
#
#  root_dir : unicode
#      A directory against against which API-style paths are to be resolved.
#
#  log : logging.Logger

## By default notebooks are saved on disk on a temporary file and then if succefully written, it replaces the old ones.
#        This procedure, namely 'atomic_writing', causes some bugs on file system whitout operation order enforcement (like some networked fs).
#        If set to False, the new notebook is written directly on the old one which could fail (eg: full filesystem or quota )
#  Default: True
# c.FileManagerMixin.use_atomic_writing = True

#------------------------------------------------------------------------------
# FileContentsManager(FileManagerMixin, ContentsManager) configuration
#------------------------------------------------------------------------------
## Allow access to hidden files
#  See also: ContentsManager.allow_hidden
# c.FileContentsManager.allow_hidden = False

## If True, deleting a non-empty directory will always be allowed.
#          WARNING this may result in files being permanently removed; e.g. on Windows,
#          if the data size is too big for the trash/recycle bin the directory will be permanently
#          deleted. If False (default), the non-empty directory will be sent to the trash only
#          if safe. And if ``delete_to_trash`` is True, the directory won't be deleted.
#  Default: False
# c.FileContentsManager.always_delete_dir = False

#  See also: ContentsManager.checkpoints
# c.FileContentsManager.checkpoints = None

#  See also: ContentsManager.checkpoints_class
# c.FileContentsManager.checkpoints_class = 'jupyter_server.services.contents.checkpoints.Checkpoints'

#  See also: ContentsManager.checkpoints_kwargs
# c.FileContentsManager.checkpoints_kwargs = {}

## If True (default), deleting files will send them to the
#          platform's trash/recycle bin, where they can be recovered. If False,
#          deleting files really deletes them.
#  Default: True
# c.FileContentsManager.delete_to_trash = True

## handler class to use when serving raw file requests.
#  See also: ContentsManager.files_handler_class
# c.FileContentsManager.files_handler_class = 'jupyter_server.files.handlers.FilesHandler'

## Extra parameters to pass to files_handler_class.
#  See also: ContentsManager.files_handler_params
# c.FileContentsManager.files_handler_params = {}

##
#  See also: ContentsManager.hide_globs
# c.FileContentsManager.hide_globs = ['__pycache__', '*.pyc', '*.pyo', '.DS_Store', '*.so', '*.dylib', '*~']

## Python callable or importstring thereof
#
#          to be called on the path of a file just saved.
#
#          This can be used to process the file on disk,
#          such as converting the notebook to a script or HTML via nbconvert.
#
#          It will be called as (all arguments passed by keyword)::
#
#              hook(os_path=os_path, model=model, contents_manager=instance)
#
#          - path: the filesystem path to the file just written
#          - model: the model representing the file
#          - contents_manager: this ContentsManager instance
#  Default: None
# c.FileContentsManager.post_save_hook = None

## Python callable or importstring thereof
#  See also: ContentsManager.pre_save_hook
# c.FileContentsManager.pre_save_hook = None

#  Default: ''
# c.FileContentsManager.root_dir = ''

## The base name used when creating untitled directories.
#  See also: ContentsManager.untitled_directory
# c.FileContentsManager.untitled_directory = 'Untitled Folder'

## The base name used when creating untitled files.
#  See also: ContentsManager.untitled_file
# c.FileContentsManager.untitled_file = 'untitled'

## The base name used when creating untitled notebooks.
#  See also: ContentsManager.untitled_notebook
# c.FileContentsManager.untitled_notebook = 'Untitled'

## By default notebooks are saved on disk on a temporary file and then if
#  succefully written, it replaces the old ones.
#  See also: FileManagerMixin.use_atomic_writing
# c.FileContentsManager.use_atomic_writing = True

#------------------------------------------------------------------------------
# AsyncContentsManager(ContentsManager) configuration
#------------------------------------------------------------------------------
## Base class for serving files and directories asynchronously.

## Allow access to hidden files
#  See also: ContentsManager.allow_hidden
# c.AsyncContentsManager.allow_hidden = False

#  Default: None
# c.AsyncContentsManager.checkpoints = None

#  Default: 'jupyter_server.services.contents.checkpoints.AsyncCheckpoints'
# c.AsyncContentsManager.checkpoints_class = 'jupyter_server.services.contents.checkpoints.AsyncCheckpoints'

#  Default: {}
# c.AsyncContentsManager.checkpoints_kwargs = {}

## handler class to use when serving raw file requests.
#  See also: ContentsManager.files_handler_class
# c.AsyncContentsManager.files_handler_class = 'jupyter_server.files.handlers.FilesHandler'

## Extra parameters to pass to files_handler_class.
#  See also: ContentsManager.files_handler_params
# c.AsyncContentsManager.files_handler_params = {}

##
#  See also: ContentsManager.hide_globs
# c.AsyncContentsManager.hide_globs = ['__pycache__', '*.pyc', '*.pyo', '.DS_Store', '*.so', '*.dylib', '*~']

## Python callable or importstring thereof
#  See also: ContentsManager.pre_save_hook
# c.AsyncContentsManager.pre_save_hook = None

#  See also: ContentsManager.root_dir
# c.AsyncContentsManager.root_dir = '/'

## The base name used when creating untitled directories.
#  See also: ContentsManager.untitled_directory
# c.AsyncContentsManager.untitled_directory = 'Untitled Folder'

## The base name used when creating untitled files.
#  See also: ContentsManager.untitled_file
# c.AsyncContentsManager.untitled_file = 'untitled'

## The base name used when creating untitled notebooks.
#  See also: ContentsManager.untitled_notebook
# c.AsyncContentsManager.untitled_notebook = 'Untitled'

#------------------------------------------------------------------------------
# AsyncFileManagerMixin(FileManagerMixin) configuration
#------------------------------------------------------------------------------
## Mixin for ContentsAPI classes that interact with the filesystem
#  asynchronously.

## By default notebooks are saved on disk on a temporary file and then if
#  succefully written, it replaces the old ones.
#  See also: FileManagerMixin.use_atomic_writing
# c.AsyncFileManagerMixin.use_atomic_writing = True

#------------------------------------------------------------------------------
# AsyncFileContentsManager(FileContentsManager, AsyncFileManagerMixin, AsyncContentsManager) configuration
#------------------------------------------------------------------------------
## Allow access to hidden files
#  See also: ContentsManager.allow_hidden
# c.AsyncFileContentsManager.allow_hidden = False

## If True, deleting a non-empty directory will always be allowed.
#  See also: FileContentsManager.always_delete_dir
# c.AsyncFileContentsManager.always_delete_dir = False

#  See also: AsyncContentsManager.checkpoints
# c.AsyncFileContentsManager.checkpoints = None

#  See also: AsyncContentsManager.checkpoints_class
# c.AsyncFileContentsManager.checkpoints_class = 'jupyter_server.services.contents.checkpoints.AsyncCheckpoints'

#  See also: AsyncContentsManager.checkpoints_kwargs
# c.AsyncFileContentsManager.checkpoints_kwargs = {}

## If True (default), deleting files will send them to the
#  See also: FileContentsManager.delete_to_trash
# c.AsyncFileContentsManager.delete_to_trash = True

## handler class to use when serving raw file requests.
#  See also: ContentsManager.files_handler_class
# c.AsyncFileContentsManager.files_handler_class = 'jupyter_server.files.handlers.FilesHandler'

## Extra parameters to pass to files_handler_class.
#  See also: ContentsManager.files_handler_params
# c.AsyncFileContentsManager.files_handler_params = {}

##
#  See also: ContentsManager.hide_globs
# c.AsyncFileContentsManager.hide_globs = ['__pycache__', '*.pyc', '*.pyo', '.DS_Store', '*.so', '*.dylib', '*~']

## Python callable or importstring thereof
#  See also: FileContentsManager.post_save_hook
# c.AsyncFileContentsManager.post_save_hook = None

## Python callable or importstring thereof
#  See also: ContentsManager.pre_save_hook
# c.AsyncFileContentsManager.pre_save_hook = None

#  See also: FileContentsManager.root_dir
# c.AsyncFileContentsManager.root_dir = ''

## The base name used when creating untitled directories.
#  See also: ContentsManager.untitled_directory
# c.AsyncFileContentsManager.untitled_directory = 'Untitled Folder'

## The base name used when creating untitled files.
#  See also: ContentsManager.untitled_file
# c.AsyncFileContentsManager.untitled_file = 'untitled'

## The base name used when creating untitled notebooks.
#  See also: ContentsManager.untitled_notebook
# c.AsyncFileContentsManager.untitled_notebook = 'Untitled'

## By default notebooks are saved on disk on a temporary file and then if
#  succefully written, it replaces the old ones.
#  See also: FileManagerMixin.use_atomic_writing
# c.AsyncFileContentsManager.use_atomic_writing = True

#------------------------------------------------------------------------------
# NotebookNotary(LoggingConfigurable) configuration
#------------------------------------------------------------------------------
## A class for computing and verifying notebook signatures.

## The hashing algorithm used to sign notebooks.
#  Choices: any of ['sha3_384', 'sha224', 'sha512', 'blake2b', 'sha3_512', 'sha384', 'sha3_224', 'md5', 'blake2s', 'sha3_256', 'sha1', 'sha256']
#  Default: 'sha256'
# c.NotebookNotary.algorithm = 'sha256'

## The storage directory for notary secret and database.
#  Default: ''
# c.NotebookNotary.data_dir = ''

## The sqlite file in which to store notebook signatures.
#          By default, this will be in your Jupyter data directory.
#          You can set it to ':memory:' to disable sqlite writing to the filesystem.
#  Default: ''
# c.NotebookNotary.db_file = ''

## The secret key with which notebooks are signed.
#  Default: b''
# c.NotebookNotary.secret = b''

## The file where the secret key is stored.
#  Default: ''
# c.NotebookNotary.secret_file = ''

## A callable returning the storage backend for notebook signatures.
#           The default uses an SQLite database.
#  Default: traitlets.Undefined
# c.NotebookNotary.store_factory = traitlets.Undefined

#------------------------------------------------------------------------------
# GatewayMappingKernelManager(AsyncMappingKernelManager) configuration
#------------------------------------------------------------------------------
## Kernel manager that supports remote kernels hosted by Jupyter Kernel or
#  Enterprise Gateway.

## Whether to send tracebacks to clients on exceptions.
#  See also: MappingKernelManager.allow_tracebacks
# c.GatewayMappingKernelManager.allow_tracebacks = True

## White list of allowed kernel message types.
#  See also: MappingKernelManager.allowed_message_types
# c.GatewayMappingKernelManager.allowed_message_types = []

## Whether messages from kernels whose frontends have disconnected should be
#  buffered in-memory.
#  See also: MappingKernelManager.buffer_offline_messages
# c.GatewayMappingKernelManager.buffer_offline_messages = True

## Whether to consider culling kernels which are busy.
#  See also: MappingKernelManager.cull_busy
# c.GatewayMappingKernelManager.cull_busy = False

## Whether to consider culling kernels which have one or more connections.
#  See also: MappingKernelManager.cull_connected
# c.GatewayMappingKernelManager.cull_connected = False

## Timeout (in seconds) after which a kernel is considered idle and ready to be
#  culled.
#  See also: MappingKernelManager.cull_idle_timeout
# c.GatewayMappingKernelManager.cull_idle_timeout = 0

## The interval (in seconds) on which to check for idle kernels exceeding the
#  cull timeout value.
#  See also: MappingKernelManager.cull_interval
# c.GatewayMappingKernelManager.cull_interval = 300

## The name of the default kernel to start
#  See also: MultiKernelManager.default_kernel_name
# c.GatewayMappingKernelManager.default_kernel_name = 'python3'

## Timeout for giving up on a kernel (in seconds).
#  See also: MappingKernelManager.kernel_info_timeout
# c.GatewayMappingKernelManager.kernel_info_timeout = 60

## The kernel manager class.  This is configurable to allow
#  See also: AsyncMultiKernelManager.kernel_manager_class
# c.GatewayMappingKernelManager.kernel_manager_class = 'jupyter_client.ioloop.AsyncIOLoopKernelManager'

#  See also: MappingKernelManager.root_dir
# c.GatewayMappingKernelManager.root_dir = ''

## Share a single zmq.Context to talk to all my kernels
#  See also: MultiKernelManager.shared_context
# c.GatewayMappingKernelManager.shared_context = True

## Message to print when allow_tracebacks is False, and an exception occurs
#  See also: MappingKernelManager.traceback_replacement_message
# c.GatewayMappingKernelManager.traceback_replacement_message = 'An exception occurred at runtime, which is not shown due to security reasons.'

#------------------------------------------------------------------------------
# GatewayKernelSpecManager(KernelSpecManager) configuration
#------------------------------------------------------------------------------
## List of allowed kernel names.
#  See also: KernelSpecManager.allowed_kernelspecs
# c.GatewayKernelSpecManager.allowed_kernelspecs = set()

## If there is no Python kernelspec registered and the IPython
#  See also: KernelSpecManager.ensure_native_kernel
# c.GatewayKernelSpecManager.ensure_native_kernel = True

## The kernel spec class.  This is configurable to allow
#  See also: KernelSpecManager.kernel_spec_class
# c.GatewayKernelSpecManager.kernel_spec_class = 'jupyter_client.kernelspec.KernelSpec'

## Deprecated, use `KernelSpecManager.allowed_kernelspecs`
#  See also: KernelSpecManager.whitelist
# c.GatewayKernelSpecManager.whitelist = set()

#------------------------------------------------------------------------------
# GatewayClient(SingletonConfigurable) configuration
#------------------------------------------------------------------------------
## This class manages the configuration.  It's its own singleton class so that we
#      can share these values across all objects.  It also contains some helper methods
#       to build request arguments out of the various config options.

## The auth scheme, added as a prefix to the authorization token used in the HTTP headers.
#          (JUPYTER_GATEWAY_AUTH_SCHEME env var)
#  Default: None
# c.GatewayClient.auth_scheme = None

## The authorization token used in the HTTP headers. The header will be formatted
#  as:
#
#              {
#                  'Authorization': '{auth_scheme} {auth_token}'
#              }
#
#          (JUPYTER_GATEWAY_AUTH_TOKEN env var)
#  Default: None
# c.GatewayClient.auth_token = None

## The filename of CA certificates or None to use defaults.
#  (JUPYTER_GATEWAY_CA_CERTS env var)
#  Default: None
# c.GatewayClient.ca_certs = None

## The filename for client SSL certificate, if any.  (JUPYTER_GATEWAY_CLIENT_CERT
#  env var)
#  Default: None
# c.GatewayClient.client_cert = None

## The filename for client SSL key, if any.  (JUPYTER_GATEWAY_CLIENT_KEY env var)
#  Default: None
# c.GatewayClient.client_key = None

## The time allowed for HTTP connection establishment with the Gateway server.
#          (JUPYTER_GATEWAY_CONNECT_TIMEOUT env var)
#  Default: 40.0
# c.GatewayClient.connect_timeout = 40.0

## A comma-separated list of environment variable names that will be included, along with
#           their values, in the kernel startup request.  The corresponding `env_whitelist` configuration
#           value must also be set on the Gateway server - since that configuration value indicates which
#           environmental values to make available to the kernel. (JUPYTER_GATEWAY_ENV_WHITELIST env var)
#  Default: ''
# c.GatewayClient.env_whitelist = ''

## The time allowed for HTTP reconnection with the Gateway server for the first time.
#              Next will be JUPYTER_GATEWAY_RETRY_INTERVAL multiplied by two in factor of numbers of retries
#              but less than JUPYTER_GATEWAY_RETRY_INTERVAL_MAX.
#              (JUPYTER_GATEWAY_RETRY_INTERVAL env var)
#  Default: 1.0
# c.GatewayClient.gateway_retry_interval = 1.0

## The maximum time allowed for HTTP reconnection retry with the Gateway server.
#              (JUPYTER_GATEWAY_RETRY_INTERVAL_MAX env var)
#  Default: 30.0
# c.GatewayClient.gateway_retry_interval_max = 30.0

## The maximum retries allowed for HTTP reconnection with the Gateway server.
#              (JUPYTER_GATEWAY_RETRY_MAX env var)
#  Default: 5
# c.GatewayClient.gateway_retry_max = 5

## Additional HTTP headers to pass on the request.  This value will be converted to a dict.
#            (JUPYTER_GATEWAY_HEADERS env var)
#  Default: '{}'
# c.GatewayClient.headers = '{}'

## The password for HTTP authentication.  (JUPYTER_GATEWAY_HTTP_PWD env var)
#  Default: None
# c.GatewayClient.http_pwd = None

## The username for HTTP authentication. (JUPYTER_GATEWAY_HTTP_USER env var)
#  Default: None
# c.GatewayClient.http_user = None

## The gateway API endpoint for accessing kernel resources
#  (JUPYTER_GATEWAY_KERNELS_ENDPOINT env var)
#  Default: '/api/kernels'
# c.GatewayClient.kernels_endpoint = '/api/kernels'

## The gateway API endpoint for accessing kernelspecs
#  (JUPYTER_GATEWAY_KERNELSPECS_ENDPOINT env var)
#  Default: '/api/kernelspecs'
# c.GatewayClient.kernelspecs_endpoint = '/api/kernelspecs'

## The gateway endpoint for accessing kernelspecs resources
#              (JUPYTER_GATEWAY_KERNELSPECS_RESOURCE_ENDPOINT env var)
#  Default: '/kernelspecs'
# c.GatewayClient.kernelspecs_resource_endpoint = '/kernelspecs'

## The time allowed for HTTP request completion. (JUPYTER_GATEWAY_REQUEST_TIMEOUT
#  env var)
#  Default: 40.0
# c.GatewayClient.request_timeout = 40.0

## The url of the Kernel or Enterprise Gateway server where
#          kernel specifications are defined and kernel management takes place.
#          If defined, this Notebook server acts as a proxy for all kernel
#          management and kernel specification retrieval.  (JUPYTER_GATEWAY_URL env var)
#  Default: None
# c.GatewayClient.url = None

## For HTTPS requests, determines if server's certificate should be validated or not.
#          (JUPYTER_GATEWAY_VALIDATE_CERT env var)
#  Default: True
# c.GatewayClient.validate_cert = True

## The websocket url of the Kernel or Enterprise Gateway server.  If not provided, this value
#          will correspond to the value of the Gateway url with 'ws' in place of 'http'.  (JUPYTER_GATEWAY_WS_URL env var)
#  Default: None
# c.GatewayClient.ws_url = None

#------------------------------------------------------------------------------
# TerminalManager(LoggingConfigurable) configuration
#------------------------------------------------------------------------------
##

## Timeout (in seconds) in which a terminal has been inactive and ready to be culled.
#          Values of 0 or lower disable culling.
#  Default: 0
# c.TerminalManager.cull_inactive_timeout = 0

## The interval (in seconds) on which to check for terminals exceeding the
#  inactive timeout value.
#  Default: 300
# c.TerminalManager.cull_interval = 300



def protected(kw):
    return '__' + kw


def get_omega(args, require_config=False):
    from omegaml import setup, _base_config
    from omegaml.client.cloud import setup_from_config
    config_file = args.get('--config')
    # deprecated, use --local
    local_runtime = args.get('--local-runtime')
    local = args.get('--local')
    if local or local_runtime:
        _base_config.OMEGA_LOCAL_RUNTIME = True
    bucket = args.get('--bucket')
    if config_file or require_config:
        try:
            om = setup_from_config(config_file)
        except Exception as e:
            msg = (f'Config file could not be found due to {e}. Specify as --config or set '
                    'OMEGA_CONFIG_FILE env variable')
            raise ValueError(msg)
    else:
        om = setup()
    if local:
        om.runtime.mode(local=True)
    return om[bucket] if bucket else om # for speed


class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self



import numpy as np
from datetime import datetime

import json

import pandas as pd
import unittest
from marshmallow import Schema, fields, ValidationError
from numpy.testing import assert_array_equal
from sklearn.linear_model import LinearRegression

from omegaml import Omega
from omegaml.backends.virtualobj import virtualobj
from omegaml.mixins.store.extdmeta import SignatureMixin, ModelSignatureMixin, ScriptSignatureMixin
from omegaml.tests.util import OmegaTestMixin


class ExtendedMetadataMixinTests(OmegaTestMixin, unittest.TestCase):
    # TODO test model versioning v.s. change of dataset metadata

    def setUp(self):
        om = self.om = Omega()
        om.datasets.register_mixin(SignatureMixin)
        om.models.register_mixin(ModelSignatureMixin)
        om.scripts.register_mixin(ScriptSignatureMixin)
        super().clean()

    def test_link_dataset(self):
        om = self.om
        df = pd.DataFrame({'x': range(10)})
        model = LinearRegression()
        metaX = om.datasets.put(df, 'testX')
        om.models.put(model, 'mymodel')
        om.models.link_dataset('mymodel', Xname='testX', data_store=om.datasets)
        meta = om.models.metadata('mymodel')
        self.assertEqual(meta.attributes['dataset']['Xname'], 'testX')
        self.assertIsInstance(meta.attributes['dataset']['Xmeta'], dict)
        self.assertIn('kind_meta', meta.attributes['dataset']['Xmeta'])

    def test_fitpredict_implicit_dataset(self):
        om = self.om
        df = pd.DataFrame({'x': range(10)})
        df['y'] = df['x'] * 5 + 1
        model = LinearRegression()
        # fit by the runtime, triggering ModelSigantureMixin._post_fit() to record dataset info
        om.datasets.put(df, 'test')
        om.models.put(model, 'mymodel')
        om.runtime.model('mymodel').fit('test[x]', 'test[y]').get()
        meta = om.models.metadata('mymodel')
        self.assertEqual(meta.attributes['dataset']['Xname'], 'test[x]')
        self.assertEqual(meta.attributes['dataset']['Yname'], 'test[y]')
        # predict using the default dataset
        yhat = om.models.get('mymodel').predict(df[['x']])
        result = om.runtime.model('mymodel').predict('*').get()
        assert_array_equal(result, yhat)
        # re-fit using the default datasets
        om.runtime.model('mymodel').fit('*', '*').get()
        meta = om.models.metadata('mymodel')
        self.assertEqual(meta.attributes['dataset']['Xname'], 'test[x]')
        self.assertEqual(meta.attributes['dataset']['Yname'], 'test[y]')

    def test_link_script_signature(self):
        om = self.om
        df = pd.DataFrame({'x': range(10)})
        om.scripts.put(myhandler, 'myhandler')

        class MyInputSchema(Schema):
            foo = fields.String()

        class MyResultSchema(Schema):
            data = fields.Dict()

        om.scripts.link_datatype('myhandler', X=MyInputSchema, result=MyResultSchema)
        resp = om.runtime.script('myhandler').run({'foo': 'bar', }).get()
        resp = json.loads(resp)
        self.assertIn('result', resp)
        self.assertEqual({"data": {'foo': 'bar'}}, resp['result'])
        with self.assertRaises(RuntimeError) as ex:
            om.runtime.script('myhandler').run({'foo': [1], }).get()
        self.assertIn("ValidationError({'foo': ['Not a valid string.']})", str(ex.exception))

    def test_link_docs(self):
        om = self.om
        model = LinearRegression()
        df = pd.DataFrame({'x': range(10)})
        om.datasets.put(df, 'test')
        model = LinearRegression()
        meta = om.models.put(model, 'mymodel')
        om.models.link_docs('mymodel', 'this is my model')
        om.datasets.link_docs('test', 'this is a dataset')
        self.assertIn('this is my model', f'{om.models.help("mymodel")}')
        self.assertIn('this is a dataset', f'{om.datasets.help("test")}')

    def test_link_datatype(self):
        class UserSchema(Schema):
            name = fields.Str()
            email = fields.Email()
            created_at = fields.DateTime()
            birthday = fields.Date()
            data = fields.List(fields.Integer)
            number = fields.Float()
            registered = fields.Bool()

        om = self.om
        model = LinearRegression()
        om.models.put(model, 'mymodel')
        om.models.link_datatype('mymodel', X=UserSchema)
        meta = om.models.metadata('mymodel')
        om.models.validate('mymodel', X={'name': 'foo'})
        with self.assertRaises(ValidationError):
            om.models.validate('mymodel', X={'birthday': 'foo'})

    def test_swagger_generator(self):
        om = self.om
        model = LinearRegression()
        om.models.put(model, 'mymodel')
        df = pd.DataFrame({'x': range(10)})
        om.datasets.put(df, 'testX')
        om.models.link_dataset('mymodel', Xname='testX', data_store=om.datasets,
                               actions=['predict', 'fit'])
        specs = om.runtime.swagger(format='dict')
        self.assertIn('paths', specs)
        self.assertIn('/api/v1/model/mymodel/predict', specs['paths'])
        self.assertIn('/api/v1/model/mymodel/fit', specs['paths'])
        self.assertIn('/api/v1/dataset/testX', specs['paths'])

    def test_swagger_generator_as_service(self):
        om = self.om
        model = LinearRegression()
        om.models.put(model, 'mymodel')
        df = pd.DataFrame({'x': range(10)})
        om.datasets.put(df, 'testX')
        om.models.link_dataset('mymodel', Xname='testX', data_store=om.datasets,
                               actions=['predict', 'fit'])
        specs = om.runtime.swagger(format='dict', as_service=True)
        self.assertIn('paths', specs)
        self.assertIn('/api/service/mymodel', specs['paths'])
        self.assertIn('/api/service/mymodel', specs['paths'])
        self.assertIn('/api/service/testX', specs['paths'])

    def test_swagger_link_specs(self):
        # sample spec
        specs = {
            'info': {
                'title': 'omega-ml service', 'version': '1.0.0'},
            'swagger': '2.0',
            'paths': {
                '/api/service/mymodel': {
                    'post': {'summary': 'summary', 'description': 'no description', 'operationId': 'mymodel#predict#post',
                             'consumes': ['application/json'], 'produces': ['application/json'],
                             'parameters': [
                                 {'in': 'body', 'name': 'body', 'description': 'no description',
                                  'schema': {'$ref': '#/definitions/mymodel_X'}}],
                             'responses': {
                                 '200': {'description': 'no description',
                                         'schema': {'$ref': '#/definitions/GeneratedSchema'}}}}},
            },
            'definitions': {
                'mymodel_X': {
                    'type': 'object', 'properties': {'x': {'type': 'integer'}}},
                'GeneratedSchema': {'type': 'object', 'properties': {'data': {}}},
                'testX': {
                    'type': 'object', 'properties': {'x': {'type': 'integer'}}},
                'DatasetInput_testX': {
                    'type': 'object', 'properties': {
                        'data': {'type': 'array', 'items': {'$ref': '#/definitions/testX'}}, 'dtypes': {},
                        'append': {'type': 'boolean'}}}}}
        om = self.om
        # create a model
        model = LinearRegression()
        om.models.put(model, 'mymodel')
        # link spec to a model
        metas = om.models.link_swagger(specs)
        self.assertEqual(len(metas), 1)
        self.assertEqual(metas[0].name, 'mymodel')
        # reverse check, we expect to have a valid spec
        specs = om.runtime.swagger(format='dict', as_service=True)
        self.assertIn('paths', specs)
        self.assertIn('/api/service/mymodel', specs['paths'])
        self.assertIn('/api/service/mymodel', specs['paths'])
        print(specs)

    def test_meta_to_schema(self):
        om = self.om
        df = pd.DataFrame({'v_int': [0],
                           'v_int32': [np.int32(0)],
                           'v_int64': [np.int64(0)],
                           'v_float': [0.25],
                           'v_float32': [np.float32(0.25)],
                           'v_float64': [np.float64(0.25)],
                           'v_obj': [list()],
                           'v_datetime': [datetime.now()],
                           # not currently supported by OmegaStore
                           # 'v_timedelta': [datetime.now() - datetime.now()],
                           # 'v_date': [datetime.now().date()],
                           'v_string': ['foo']})
        meta = om.datasets.put(df, 'test')
        extdmeta = SignatureMixin()
        Schema = extdmeta._datatype_from_metadata(meta.to_dict())
        sfields = Schema().fields
        self.assertIsInstance(sfields['v_int'], fields.Integer)
        self.assertIsInstance(sfields['v_int32'], fields.Integer)
        self.assertIsInstance(sfields['v_int64'], fields.Integer)
        self.assertIsInstance(sfields['v_float'], fields.Float)
        self.assertIsInstance(sfields['v_float32'], fields.Float)
        self.assertIsInstance(sfields['v_float64'], fields.Float)
        self.assertIsInstance(sfields['v_obj'], fields.String)
        self.assertIsInstance(sfields['v_string'], fields.String)


@virtualobj
def myhandler(*args, **kwargs):
    return {'data': kwargs.get('data')}


if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import

import logging

logger = logging.getLogger(__file__)


class OmegaScriptProxy(object):
    """
    proxy to a remote job in a celery worker

    Usage:

        .. code::

            om = Omega()
            # result is AsyncResult, use .get() to return it's result
            result = om.runtimes.script('foojob').run()
            result.get()
    """

    def __init__(self, scriptname, runtime=None):
        self.scriptname = scriptname
        self.runtime = runtime

    def run(self, *args, as_callback=False, **kwargs):
        """
        run the script

        Runs the script and returns its result as::

            {"runtimes": 7.4e-05,
              "started": "2018-04-07T17:57:52.451012",
              "kwargs": {},
              "result": {},
              "script": "helloworld"
            }

        Returns:
            AsyncResult
        """
        script_run = self.task(as_callback=as_callback)
        return script_run.delay(self.scriptname, *args, **kwargs)

    def task(self, as_callback=False):
        task_name = 'run_omega_callback_script' if as_callback else 'run_omega_script'
        return self.runtime.task(f'omegaml.backends.package.tasks.{task_name}')



import logging
import os
import requests
import subprocess
from pathlib import Path
from pprint import pprint
from subprocess import call
from time import sleep

from omegaml.client.cli import deploy
from omegaml.client.docoptparser import CommandBase
from omegaml.client.util import get_omega
from omegaml.mixins.store.imexport import OmegaExporter


class RuntimeCommandBase(CommandBase):
    """
    Usage:
      om runtime model <name> <model-action> [<X>] [<Y>] [--result=<output-name>] [--param=<kw=value>]... [--async] [options]
      om runtime script <name> [<script-action>] [<kw=value>...] [--async] [options]
      om runtime job <name> [<job-action>] [<args...>] [--async] [options]
      om runtime result <taskid> [options]
      om runtime ping [options]
      om runtime env <action> [<package>] [--file <requirements.txt>] [--every] [options]
      om runtime log [-f] [options]
      om runtime status [workers|labels|stats] [options]
      om runtime restart app <name> [--insecure] [--apphub-url=<url>] [options]
      om runtime serve [<rule>...] [--rules=<rulefile>] [options]
      om runtime deploy [<deploy-action>] [--steps=<deployfile>] [--specs=<specs>] [--select=<filter>] [--dry] [options]
      om runtime (control|inspect|celery) [<celery-command>...] [--worker=<worker>] [--queue=<queue>] [--celery-help] [--flags <celery-flags>] [options]
      om runtime (export|import) [<prefix/name>...] [--path=<path>] [--compress] [--list] [--promote] [options]

      Options:
      --async             don't wait for results, will print taskid
      -f                  tail log
      --require=VALUE     worker label
      --flags=VALUES      celery flags, e.g. --flags "--queues=A,B,C -E --loglevel=INFO"
      --worker=VALUE      celery worker
      --queue=VALUE       celery queue
      --celery-help       show celery help
      --file=VALUE        path/to/requirements.txt
      --local             if specified the task will run locally. Use this for testing
      --every             if specified runs task on all workers
      --insecure          allow insecure connections (disables verification of ssl certificates)
      --rules=VALUE       /path/to/specs.txt, where each line is a <spec>
      --compress          if specified the archive will be compress (tgz format) [default: True]
      --path=PATH         path to directory where the archive should be written [default: ./mlops-export]
      --list              if specified, print members of archive
      --promote           if specified, import and promote objects
      --steps=VALUE       /path/to/deployfile.yaml [default: ./deployfile.yaml]
      --apphub-url=VALUE  the url of the apphub instance

    Description:
      model, job and script commands
      ------------------------------

      <model-action> can be any valid model action like fit, predict, score,
      transform, decision_function etc.

      <script-action> defaults to run
      <job-action> defaults to run

      Examples:
        om runtime model <name> fit <X> <Y>
        om runtime model <name> predict <X>
        om runtime job <name>
        om runtime script <name>
        om runtime script <name> run myparam="value"

      running asynchronously
      ----------------------

      model, job, script commands accept the --async paramter. This will submit
      the a task and return the task id. To wait for and get the result run use
      the result command

      Examples:
            om runtime model <name> fit <X> <Y> --async
            => <task id>
            om runtime result <task id>
            => result of the task

      restart app
      -----------

      This will restart the app on omegaml apphub. Requires a login to omegaml cloud.


      status
      ------

      Prints workers, labels, list of active tasks per worker, count of tasks

      Examples:
        om runtime status             # defaults to workers
        om runtime status workers
        om runtime status labels
        om runtime status stats

      celery commands
      ---------------

      This is the same as calling celery -A omegaml.celeryapp <commands>. Command
      commands include:

      inspect active         show currently running tasks
      inspect active_queues  show active queues for each worker
      inspect stats          show stats of each worker, including pool size (processes)
      inspect ping           confirm that worker is connected

      control pool_shrink N  shrink worker pool by N, specify 99 to remove all
      control pool_grow N    grow worker poool by N
      control shutdown       stop and restart the worker

      Examples:
            om runtime celery inspect active
            om runtime celery control pool_grow N


      env commands
      ------------

      This talks to an omegaml worker's pip environment

      a) install a specific package

         env install <package>    install the specified package, use name==version pip syntax for specific versions
         env uninstall <package>  uninstall the specified package

         <package> is in pip install syntax, e.g.

         env install "six==1.0.0"
         env install "git+https://github.com/user/repo.git"

      b) use a requirements file

         env install --file requirements.txt
         env uninstall --file requirements.txt

      c) list currently installed packages

         env freeze
         env list

      d) install on all or a specific worker

         env install --require gpu package
         env install --every package

         By default the installation runs on the default worker only. If there are multiple nodes where you
         want to install the package(s) worker nodes, be sure to specify --every

      Examples:
            om runtime env install pandas
            om runtime env uninstall pandas
            om runtime env install --file requirements.txt
            om runtime env install --file gpu-requirements.txt --require gpu
            om runtime env install --file requirements.txt --every

      serve command
      -------------

      Start a local omega-ml server for models, scripts, jobs. Each <rule> is
      a regex in format <resource>/<name>/<action>. If you provide a <rulefile>,
      each line is a <rule>, lines that start by # are ignored.

      * resource: model|script|job
      * name: name or pattern of existing objects
      * action: resource actions

      Examples:
          # allow all models, scripts, jobs
          om runtime serve
          om runtime serve .*/.*/.*

          # allow all models, all actions, but no other resources
          om runtime serve model/.*/.*

          # allow only model foo and dataset xyz
          om runtime serve "model/foo/.*" "dataset/xyz/.*"

          # allow only predictions by model foo
          om runtime serve model/foo/predict


      deploy command
      --------------

      To deploy a complete application, create the deploy.yml as per below
      and run:

        # initial deployment
        $ om runtime deploy --dry

        # subsequent deployments
        $ om runtime deploy

        # select specific parts to deploy
        $ om runtime deploy --select scripts.apps/helloworld

        # dry run, show commands, do not exeucte
        $ om runtime deploy --dry

        # to see an example of the configuration file
        $ om runtime deploy example
    """
    command = 'runtime'

    def ping(self):
        om = get_omega(self.args)
        label = self.args.get('--require')
        self.logger.info(om.runtime.require(label).ping())

    def _ensure_valid_XY(self, value):
        if value is not None:
            if value.isnumeric():
                return [eval(value)]
            if value[0] == '[' and value[-1] == ']':
                return eval(value)
        return value

    def model(self):
        om = get_omega(self.args)
        name = self.args.get('<name>')
        action = self.args.get('<model-action>')
        is_async = self.args.get('--async')
        kwargs_lst = self.args.get('--param')
        output = self.args.get('--result')
        label = self.args.get('--require')
        X = self._ensure_valid_XY(self.args.get('<X>'))
        Y = self._ensure_valid_XY(self.args.get('<Y>'))
        # parse the list of kw=value values
        # e.g. key1=val1 key2=val2 => kwargs_lst = ['key1=val1', 'key2=val2']
        #   => kw_dct = { 'key1': eval('val1'), 'key2': eval('val2') }
        kv_dct = {}
        for kv in kwargs_lst:
            k, v = kv.split('=', 1)
            kv_dct[k] = eval(v)
        kwargs = {}
        if action in ('predict', 'predict_proba',
                      'decision_function', 'transform'):
            # actions that take rName, but no Y
            kwargs['rName'] = output
        else:
            # actions that take Y, but no rName
            kwargs['Yname'] = Y
        if action == 'gridsearch':
            kwargs['parameters'] = kv_dct
        rt_model = om.runtime.require(label).model(name)
        meth = getattr(rt_model, action, None)
        if meth is not None:
            result = meth(X, **kwargs)
            if not is_async:
                self.logger.info(result.get())
            else:
                self.logger.info(result)
            return
        raise ValueError('{action} is not applicable to {name}'.format(**locals()))

    def script(self):
        om = get_omega(self.args)
        name = self.args.get('<name>')
        is_async = self.args.get('--async')
        kwargs = self.parse_kwargs('<kw=value>')
        label = self.args.get('--require')
        result = om.runtime.require(label).script(name).run(**kwargs)
        if not is_async:
            self.logger.info(result.get())
        else:
            self.logger.info(result.task_id)

    def job(self):
        om = get_omega(self.args)
        name = self.args.get('<name>')
        is_async = self.args.get('--async')
        label = self.args.get('--require')
        result = om.runtime.require(label).job(name).run()
        if not is_async:
            self.logger.info(result.get())
        else:
            self.logger.info(result.task_id)

    def result(self):
        from celery.result import AsyncResult

        om = get_omega(self.args)
        task_id = self.args.get('<taskid>')
        result = AsyncResult(task_id, app=om.runtime.celeryapp).get()
        self.logger.info(result)

    def log(self):
        import pandas as pd
        tail = self.args.get('-f')
        om = get_omega(self.args)
        if not tail:
            df = om.logger.dataset.get()
            with pd.option_context('display.max_rows', None,
                                   'display.max_columns', None,
                                   'display.max_colwidth', -1):
                print(df[['text']])
        else:
            om.logger.dataset.tail()

    def inspect(self):
        self.celery('inspect')

    def control(self):
        self.celery('control')

    def celery(self, action=None):
        om = get_omega(self.args)
        # giving om command here changes celery help output
        celery_cmds = ['om runtime celery']
        if action:
            celery_cmds += action.split(' ')
        # convert omega terms into celery terms
        celery_opts = (
            # omega term, celery term, value|flag
            ('--worker', '--destination', 'value'),
            ('--queue', '--queue', 'value'),
            ('--celery-help', '--help', 'flag'),
        )
        is_r_worker = 'rworker' in self.args.get('<celery-command>')
        for opt, celery_opt, kind in celery_opts:
            if self.args.get(opt):
                celery_cmds += [celery_opt]
                if kind == 'value':
                    celery_cmds += [self.args.get(opt)]
        # prepare celery command args, remove empty parts
        celery_cmds += self.args.get('<celery-command>')
        celery_cmds += (self.args.get('--flags') or '').split(' ')
        celery_cmds = [cmd for cmd in celery_cmds if cmd]
        if len(celery_cmds) == 1 + int(action is not None):
            celery_cmds += ['--help']
        # start in-process for speed
        # -- disable command logging to avoid curses problems in celery events
        self.logger.setLevel(logging.CRITICAL + 1)
        if is_r_worker:
            # start r runtime
            from omegaml.runtimes import rsystem
            rworker = os.path.join(os.path.dirname(rsystem.__file__), 'omworker.R')
            rcmd = f'Rscript {rworker}'.split(' ')
            call(rcmd)
        else:
            # start python runtime
            om.runtime.celeryapp.start(celery_cmds)

    def env(self):
        om = get_omega(self.args)
        action = self.args.get('<action>')
        package = self.args.get('<package>')
        reqfile = self.args.get('--file')
        every = self.args.get('--every')
        require = self.args.get('--require') or ''
        if reqfile:
            with open(reqfile, 'rb') as fin:
                om.scripts.put(fin, '.system/requirements.txt')
        if not om.scripts.exists('.system/envinstall', hidden=True):
            import omegaml as om_module
            envinstall_path = os.path.join(os.path.dirname(om_module.__file__), 'runtimes', 'envinstall')
            om.scripts.put(f'pkg://{envinstall_path}', '.system/envinstall')
        if every:
            labels = om.runtime.enable_hostqueues()
        else:
            labels = require.split(',')
        results = []
        for label in labels:
            result = (om.runtime.require(label)
                      .script('.system/envinstall')
                      .run(action=action, package=package, file=reqfile,
                           __format='python'))
            results.append((label, result))
        all_results = om.runtime.celeryapp.ResultSet([r[1] for r in results])
        from tqdm import tqdm
        with tqdm() as progress:
            while all_results.waiting():
                progress.update(1)
                sleep(1)
            all_results.get()
        for label, result in results:
            if label:
                print(f'** result of worker require={label}:')
            data = result.get()  # resolve AsyncResult => dict
            print(str(data.get('result', data)))  # get actual result object, pip stdout

    def status(self):
        om = get_omega(self.args)
        labels = self.args.get('labels')
        stats = self.args.get('stats')
        workers = not (labels or stats)
        if workers:
            pprint(om.runtime.workers())
        elif labels:
            queues = om.runtime.queues()
            pprint({worker: [q.get('name') for q in details
                             if not q.get('name').startswith('amq')]
                    for worker, details in queues.items()})
        elif stats:
            stats = om.runtime.stats()
            pprint({worker: {
                'size': details['pool']['max-concurrency'],
                'tasks': {
                    task: count for task, count in details['total'].items()
                }
            } for worker, details in stats.items()})

    def restart(self):
        from omegaml.client.userconf import ensure_api_url

        om = get_omega(self.args)
        name = self.args.get('<name>')
        insecure = self.args.get('--insecure', False)
        apphub_url = self.args.get('--apphub-url')
        user = om.runtime.auth.userid
        qualifier = om.runtime.auth.qualifier
        auth = requests.auth.HTTPBasicAuth(user, om.runtime.auth.apikey)
        headers = {'Qualifier': qualifier}
        appsurl = ensure_api_url(apphub_url, om.defaults, key='OMEGA_APPHUB_URL')
        name = name.replace('apps/', '')
        stop = requests.get(f'{appsurl}/apps/api/stop/{user}/{name}'.format(om.runtime.auth.userid),
                            auth=auth, verify=not insecure, headers=headers)
        start = requests.get(f'{appsurl}/apps/api/start/{user}/{name}'.format(om.runtime.auth.userid),
                             auth=auth, verify=not insecure, headers=headers)
        self.logger.info(f'stop: {stop} start: {start}')

    def serve(self):
        om = get_omega(self.args, require_config=False)
        specs = self.args.get('<rule>')
        specfile = self.args.get('--rules')
        if specfile:
            with open(specfile, 'r') as fin:
                specs = [s.replace('\n', '') for s in fin.readlines() if not s.startswith('#')]
        os.environ['OMEGA_RESTAPI_FILTER'] = ';'.join(specs) if specs else om.defaults.OMEGA_RESTAPI_FILTER
        subprocess.run("gunicorn 'omegaml.restapi.app:serve_objects()'", shell=True)

    def deploy(self):
        om = get_omega(self.args, require_config=False)
        deployfile = self.args.get('--steps') or 'deployfile.yaml'
        action = self.args.get('<deploy-action>') or 'update'
        dry = self.args.get('--dry')
        specs = self.args.get('--specs')
        selection = self.args.get('<filter>')
        if action == 'example':
            help(deploy)
        assert Path(deployfile).exists(), f'{deployfile} does not exist'
        deploy.process(deployfile, action=action, dry=dry, specs=specs, select=selection, cli_logger=self.logger)

    def do_export(self):
        om = get_omega(self.args)
        names = self.args.get('<prefix/name>')
        archive = self.args.get('--path') or './mlops-export'
        compress = self.args.get('--compress', True)
        dolist = self.args.get('--list')
        if dolist:
            arc = OmegaExporter.archive(archive)
            self.print(list(arc.members))
        else:
            exp = OmegaExporter(om)
            arcfile = exp.to_archive(archive, names, compress=compress, progressfn=print)
            self.print(arcfile)

    def do_import(self):
        om = get_omega(self.args)
        names = self.args.get('<prefix/name>')
        archive = Path(self.args.get('--path', './mlops-export'))
        dolist = self.args.get('--list')
        promote = self.args.get('--promote')
        pattern = '|'.join(names)
        if (not archive.is_file()
              and not archive.exists()
              and archive.parent.exists()):
            archives = list(archive.parent.glob(f'{archive.name}*'))
            if len(archives) > 1:
                archive = Path(self.ask("Select an archive", options=archives, select=True, default=1))
        assert archive.exists(), f"No mlops-export {archive} exists"
        self.print(f"Processing {archive}")
        if dolist:
            arc = OmegaExporter.archive(archive)
            self.print(list(arc.members))
        else:
            exp = OmegaExporter(om)
            arcfile = exp.from_archive(archive, pattern=pattern, progressfn=print, promote=promote)
            self.print("Imported objects:")
            self.print(arcfile)



import os
from mongoengine import GridFSProxy
from warnings import warn


class BackendBaseCommon:
    """
    common base for storage backends
    """
    def _tmp_packagefn(self, store, name):
        """
        use this to to get a temporary local filename for serialization/deserialization

        This uses the store's .tmppath to generate a valid fully qualified path, ensuring
        the directory exists. Note that you must clean up files yourself after use, i.e.
        call os.remove()

        Args:
            store (OmegaStore): the store to get the temporary path
            name (str): the name of the file

        Returns:
            filename (str)

        TODO: register all files to remove temp files on process exit
        """
        filename = os.path.join(store.tmppath, name)
        dirname = os.path.dirname(filename)
        os.makedirs(dirname, exist_ok=True)
        return filename

    def _is_path(self, obj):
        return isinstance(obj, str) and os.path.exists(obj)

    def _store_to_file(self, store, obj, filename, encoding=None, replace=False):
        """
        Use this method to store file-like objects to the store's gridfs

        Args:
            store (OmegaStore): the store whose .fs filesystem access will be used
            obj (file-like): a file-like object or path, if path will be opened with mode=rb,
               otherwise its obj.read() method is called to get the data as a byte stream
            filename (path): the path in the store (key)
            encoding (str): a valid encoding such as utf8, optional
            replace (bool): if True the existing file(s) of the same name are deleted to avoid automated versioning
               by gridfs. defaults to False

        Returns:
            gridfile (GridFSProxy), assignable to Metadata.gridfile
        """
        if replace:
            for fileobj in store.fs.find({'filename': filename}):
                try:
                    store.fs.delete(fileobj._id)
                except Exception as e:
                    warn('deleting {filename} resulted in {e}'.format(**locals()))
                    pass
        if self._is_path(obj):
            with open(obj, 'rb') as fin:
                fileid = store.fs.put(fin, filename=filename, encoding=encoding)
        else:
            fileid = store.fs.put(obj, filename=filename, encoding=encoding)
        gridfile = GridFSProxy(grid_id=fileid,
                               db_alias=store._dbalias,
                               key=filename,
                               collection_name=store._fs_collection)
        return gridfile

    def perform(self, method, *args, **kwargs):
        """ perform a model action, wrapped by pre-action/post-action calls

        This is a helper method for the OmegaRuntime tasks to call model
        actions that require pre/post processing. The pre/post action methods
        are looked up on the store provided by the Backend._call_handler,
        enabling mixins to the model store to handle such calls.

        Usage:
            backend.perform('action', *args, **kwargs)

            => will call

                args, kwargs = handler.pre_<action>(*args, **kwargs),
                result = backend.<action>(**args, **kwargs)
                handler.post_<action>(result, *args, **kwargs)

                where <action:str> is the effective method to be called,
                handler is the _call_handler property of the object's backend,
                typically the store where the object is contained.

                The pre_<action> method should be defined as::

                    def pre_action(**args, **kwargs):
                        ...
                        return args, kwargs

                The args, kwargs will be passed as action(*args, **kwargs)
                to the effective method. Note you can use any definition of
                *args, **kwargs as long as you return a tuple([list], dict).

                The post_<action> method should be defined as::

                    def post_action(result, *args, **kwargs):
                        ...
                        return result

                The args, kwargs are the same as passed to the pre_<action>()
                and action() methods. result is the return value from the
                effective method and should be the only value returned from
                the post_<action>() method (though it can be modified as you
                see fit).

        Notes:

            - see the ModelSignatureMixin for pre/post action methods on fit()
              and predict()
            - the rationale to use store mixins for pre and post action is that
              this are generic across all backends. However the backend may
              override by providing its own _call_handler (which by default is
              the main store of the backend (i.e. model_store for models, data_store
              for datasets, scripts_store for scripts etc.)
        """
        pre_nop = lambda *args, **kwargs: (args, kwargs)
        post_nop = lambda v, *args, **kwargs: v
        do_call = getattr(self, method, None)
        pre_call = getattr(self._call_handler, f'_pre_{method}', pre_nop)
        post_call = getattr(self._call_handler, f'_post_{method}', post_nop)
        common_kwargs = dict(data_store=getattr(self, 'data_store'),
                             model_store=getattr(self, 'model_store'))
        args, kwargs = pre_call(*args, **kwargs)
        result = do_call(*args, **kwargs)
        return post_call(result, *args, **kwargs, **common_kwargs)

    @property
    def _call_handler(self):
        # by default the data store handles _pre and _post methods in self.perform()
        return self.data_store





from io import StringIO
from unittest import TestCase
from unittest.mock import patch

from omegaml import Omega
from omegaml.client.auth import AuthenticationEnv
from omegaml.defaults import update_from_config, update_from_obj, update_from_dict


class BareObj(object):
    pass


class ConfigurationTests(TestCase):
    def setUp(self):
        self.defaults = {
            'OMEGA_MONGO_URL': 'foo'
        }

    def tearDown(self):
        # reset authentaction env to avoid interference with other tests
        AuthenticationEnv.auth_env = None

    def test_update_from_configfile(self):
        """
        Test updates from config file works
        """
        data = """
        OMEGA_MONGO_URL: updated-foo
        NOTOMEGA_VALUE: some other value
        """
        update_from_config(vars=self.defaults, config_file=StringIO(data))
        self.assertIn('OMEGA_MONGO_URL', self.defaults)
        self.assertEqual(self.defaults['OMEGA_MONGO_URL'], 'updated-foo')
        self.assertNotIn('NOTOMEGA_VALUE', self.defaults)

    def test_update_from_obj(self):
        data = BareObj()
        data.OMEGA_MONGO_URL = 'updated-foo'
        update_from_obj(data, self.defaults)
        self.assertIn('OMEGA_MONGO_URL', self.defaults)
        self.assertEqual(self.defaults['OMEGA_MONGO_URL'], 'updated-foo')
        self.assertNotIn('NOTOMEGA_VALUE', self.defaults)

    def test_update_from_obj_nested(self):
        data = BareObj()
        data.OMEGA_CELERY_CONFIG = {
            'TASK_SERIALIZER': 'json',
        }
        self.defaults['OMEGA_CELERY_CONFIG'] = {
            'RESULT_SERIALIZER': 'pickle',
            'TASK_SERIALIZER': 'pickle',
        }
        update_from_obj(data, self.defaults)
        self.assertEqual(self.defaults['OMEGA_CELERY_CONFIG']['RESULT_SERIALIZER'], 'pickle')
        self.assertEqual(self.defaults['OMEGA_CELERY_CONFIG']['TASK_SERIALIZER'], 'json')

    def test_update_obj_from_obj_nested(self):
        data = BareObj()
        data.OMEGA_CELERY_CONFIG = {
            'TASK_SERIALIZER': 'json',
        }
        defaults = BareObj()
        defaults.OMEGA_CELERY_CONFIG = {
            'RESULT_SERIALIZER': 'pickle',
            'TASK_SERIALIZER': 'pickle',
        }
        update_from_obj(data, defaults)
        self.assertEqual(defaults.OMEGA_CELERY_CONFIG['RESULT_SERIALIZER'], 'pickle')
        self.assertEqual(defaults.OMEGA_CELERY_CONFIG['TASK_SERIALIZER'], 'json')

    def test_update_from_dict(self):
        data = {
            'OMEGA_MONGO_URL': 'updated-foo'
        }
        update_from_dict(data, self.defaults)
        self.assertIn('OMEGA_MONGO_URL', self.defaults)
        self.assertEqual(self.defaults['OMEGA_MONGO_URL'], 'updated-foo')
        self.assertNotIn('NOTOMEGA_VALUE', self.defaults)

    def test_config_from_apikey(self):
        """
        Test an Omega instance can be created from user specific configs
        """
        import omegaml as om
        from omegaml.util import settings
        # check we get default without patching
        from omegaml import _base_config as _real_base_config
        with patch('omegaml._base_config', new=BareObj) as defaults:
            # link callbacks used by get_omega_from_api_key
            _real_base_config.update_from_obj(_real_base_config, attrs=defaults)
            defaults.update_from_dict = _real_base_config.update_from_dict
            defaults.update_from_config = _real_base_config.update_from_config
            defaults.load_config_file = _real_base_config.load_config_file
            defaults.update_from_env = _real_base_config.update_from_env
            defaults.load_user_extensions = lambda *args, **kwargs: None
            defaults.load_framework_support = lambda *args, **kwargs: None
            setup = om.setup
            defaults.MY_OWN_SETTING = 'foo'
            settings(reload=True)
            om = om.setup()
            self.assertEqual(om.defaults.MY_OWN_SETTING, 'foo')
            # reset om.datasets to restored defaults
            om = setup()
            self.assertNotEqual(om.datasets.mongo_url, 'foo')
        # now test we can change the default through config
        # we patch the actual api call to avoid having to set up the user db
        # the objective here is to test get_omega_from_apikey
        with patch('omegaml.client.userconf._get_userconfig_from_api') as mock:
            mock.return_value = {
                'objects': [
                    {
                        'data': {
                            'OMEGA_MONGO_URL': 'updated-foo',
                            'OMEGA_MY_OWN_SETTING': 'updated-foo',
                            'OMEGA_CELERY_CONFIG': {
                                'TEST_SETTING': 'pickle',
                            }
                        }
                    }
                ]
            }
            with patch('omegaml._base_config', new=BareObj) as defaults:
                from omegaml.client.userconf import get_omega_from_apikey
                # link callbacks used by get_omega_from_api_key
                _real_base_config.update_from_obj(_real_base_config, attrs=defaults)
                defaults.update_from_dict = _real_base_config.update_from_dict
                defaults.load_user_extensions = lambda *args, **kwargs: None
                defaults.load_framework_support = lambda *args, **kwargs: None
                defaults.OMEGA_MY_OWN_SETTING = 'foo'
                om = get_omega_from_apikey('foo', 'bar')
                self.assertEqual(om.defaults.OMEGA_MY_OWN_SETTING, 'updated-foo')
                self.assertEqual(om.datasets.mongo_url, 'updated-foo')
                self.assertEqual(om.defaults.OMEGA_CELERY_CONFIG['TEST_SETTING'], 'pickle')
                # test that all default values are still there, i.e. the OMEGA_CELERY_CONFIG was updated, not replaced
                for real_k, real_v in _real_base_config.OMEGA_CELERY_CONFIG.items():
                    self.assertIn(real_k, om.defaults.OMEGA_CELERY_CONFIG)
        # restore defaults
        defaults = settings(reload=True)
        om = setup()
        self.assertNotEqual(om.datasets.mongo_url, 'foo')

    def test_user_extensions_config(self):
        # check we get default without patching
        from omegaml.util import settings
        from omegaml import _base_config as _real_base_config
        with patch('omegaml.client.userconf._get_userconfig_from_api') as mock:
            mock.return_value = {
                'objects': [
                    {
                        'data': {
                            "OMEGA_USER_EXTENSIONS": {
                                "OMEGA_STORE_BACKENDS": {
                                    "test.backend": 'omegaml.backends.npndarray.NumpyNDArrayBackend'
                                }}}}]
            }
            with patch('omegaml._base_config', new=BareObj) as defaults:
                from omegaml.client.userconf import get_omega_from_apikey
                # link callbacks used by get_omega_from_api_key
                _real_base_config.update_from_obj(_real_base_config, attrs=defaults)
                defaults.update_from_dict = _real_base_config.update_from_dict
                defaults.update_from_config = _real_base_config.update_from_config
                defaults.load_user_extensions = _real_base_config.load_user_extensions
                defaults.load_config_file = _real_base_config.load_config_file
                defaults.update_from_env = _real_base_config.update_from_env
                defaults.load_framework_support = lambda *args, **kwargs: None
                defaults.OMEGA_MY_OWN_SETTING = 'foo'
                om = get_omega_from_apikey('foo', 'bar')
                self.assertIsNotNone(om.defaults.OMEGA_USER_EXTENSIONS)
                self.assertIn('test.backend', om.defaults.OMEGA_STORE_BACKENDS)
                # test that all default values are still there, i.e. config was updated, not replaced
                for real_k, real_v in _real_base_config.OMEGA_STORE_BACKENDS.items():
                    self.assertIn(real_k, om.defaults.OMEGA_STORE_BACKENDS)
                # restore defaults
            defaults = settings(reload=True)

    def test_defaults_repr(self):
        # ensure accessing defaults in shell does not cause RecursionError
        om = Omega()
        not_raised = False
        try:
            context_repr = repr(om.defaults)
        except RecursionError as e:
            pass
        else:
            not_raised = True
        self.assertTrue(not_raised)



import json

import warnings

import os
import sys
import yaml
from urllib3 import Retry

from omegaml.client.auth import AuthenticationEnv

default_key = 'OMEGA_RESTAPI_URL'
fallback_url = 'https://hub.omegaml.io'


def ensure_api_url(api_url, defaults, key=default_key, fallback=fallback_url):
    api_url_default = os.environ.get(key) or os.environ.get(default_key) or fallback
    api_url = api_url or getattr(defaults, key, getattr(defaults, default_key, api_url_default))
    return api_url


def session_backoff(retries=5):
    import requests
    from requests.adapters import HTTPAdapter
    s = requests.Session()
    retry = Retry(total=retries,
                  backoff_factor=.1,
                  status_forcelist=[500, 502, 503, 504])
    s.mount('http://', HTTPAdapter(max_retries=retry))
    s.mount('https://', HTTPAdapter(max_retries=retry))
    return s


def _get_userconfig_from_api(api_auth, api_url=None, requested_userid=None, qualifier=None, view=False):
    # safe way to talk to either the remote API or the in-process test server
    from omegaml import settings
    defaults = settings()
    api_url = ensure_api_url(api_url, defaults)
    api_url += '/api/v1/config/'
    api_url = api_url.replace('//api', '/api')
    query = []
    if requested_userid:
        query.append('user={}'.format(requested_userid))
    if view:
        query.append('view={}'.format(int(view)))
    if qualifier:
        query.append('qualifier={}'.format(qualifier))
    api_url += '?' + '&'.join(query)
    # -- setup appropriate client API
    if api_url.startswith('http'):
        import requests
        server = session_backoff()
        server_kwargs = dict(auth=api_auth)
        deserialize = lambda resp: resp.json()
    elif api_url.startswith('test') or any('test' in v for v in sys.argv):
        try:
            from tastypie.test import TestApiClient
        except ModuleNotFoundError as e:
            # we need omegaee environment to proceed
            raise
        server = TestApiClient()
        server.close = lambda: None
        server_kwargs = dict(authentication=api_auth.get_credentials())
        deserialize = lambda resp: json.loads(resp.content.decode('utf-8'))
    else:
        raise ValueError('invalid api_url >{}<'.format(api_url))
    # -- actual logic to get configs
    fail_msg = ("omegaml hub refused authentication by {api_auth}, "
                "status code={resp.status_code} "
                "using {api_url}.")
    resp = server.get(api_url, **server_kwargs)
    assert resp.status_code == 200, fail_msg.format(**locals())
    configs = deserialize(resp)
    server.close()
    return configs


# FIXME enable cache by arguments (mnemonic) @cached(seconds=3600)
def get_omega_from_apikey(*args, **kwargs):
    warnings.warn(('Calling get_omega_from_apikey directly is deprecated '
                   'and will be removed in the next release. Please use  '
                   'AuthenticationEnv.secure().get_omega_from_apikey'), DeprecationWarning)
    return _get_omega_from_apikey(*args, **kwargs)


def _get_omega_from_apikey(userid, apikey, api_url=None, requested_userid=None,
                           qualifier=None, view=False):
    """
    setup an Omega instance from userid and apikey

    :param userid: the userid
    :param apikey: the apikey
    :param api_url: the api URL
    :param requested_userid: the userid to request config for. in this case userid
      and apikey must for a staff user for the request to succeed
    :param qualifier: the database qualifier requested. defaults to 'default'
    :returns: OmegaCloud instance configured for the given user

    Returns:
        OmegaCloud
    """
    from omegaml.client.cloud import OmegaCloud
    from omegaml import settings, _base_config

    defaults = settings(reload=True)
    qualifier = qualifier or 'default'
    defaults.OMEGA_USERID = userid
    defaults.OMEGA_APIKEY = apikey
    defaults.OMEGA_QUALIFIER = qualifier
    api_url = ensure_api_url(api_url, defaults)
    auth_env = AuthenticationEnv.secure()
    if api_url.startswith('http') or any('test' in v for v in sys.argv):
        configs = auth_env.get_userconfig_from_api(requested_userid=requested_userid,
                                                   api_url=api_url, view=view,
                                                   defaults=defaults)
        configs = configs['objects'][0]['data']
    elif api_url == 'local':
        configs = {k: getattr(defaults, k) for k in dir(defaults) if k.startswith('OMEGA')}
    else:
        raise ValueError('invalid api_url >{}<'.format(api_url))
    config = configs.get(qualifier, configs)
    # update
    _base_config.update_from_dict(config, attrs=defaults)
    _base_config.update_from_config(defaults)
    _base_config.update_from_env(defaults)
    _base_config.load_framework_support(defaults)
    _base_config.load_user_extensions(defaults)
    # update config to reflect request
    defaults.OMEGA_RESTAPI_URL = api_url
    defaults.OMEGA_USERID = userid
    defaults.OMEGA_APIKEY = apikey
    defaults.OMEGA_QUALIFIER = qualifier
    auth = auth_env.get_runtime_auth(defaults)
    om = OmegaCloud(defaults=defaults, auth=auth)
    return om


def get_omega_from_config(*args, **kwargs):
    warnings.warn(('Calling _get_omega_from_config directly is deprecated '
                   'and will be removed in the next release. Please use  '
                   'AuthenticationEnv.secure()._get_omega_from_config'), DeprecationWarning)
    return _get_omega_from_config(*args, **kwargs)


def _get_omega_from_config(configfile, qualifier=None):
    from omegaml import Omega
    from omegaml import settings, _base_config
    defaults = settings()
    with open(configfile, 'r') as fconfig:
        configs = yaml.safe_load(fconfig)
    qualifier = qualifier or 'default'
    if qualifier == 'default':
        config = configs.get(qualifier, configs)
    else:
        config = configs[qualifier]
    _base_config.update_from_dict(config, attrs=defaults)
    settings(reload=True)
    om = Omega(defaults=defaults)
    return om


def _save_userconfig_from_apikey(configfile, userid, apikey, api_url=None, requested_userid=None,
                                 view=False, keys=None, qualifier=None):
    from omegaml import settings
    defaults = settings()
    api_url = ensure_api_url(api_url, defaults)
    required_keys = ['OMEGA_USERID', 'OMEGA_APIKEY', 'OMEGA_RESTAPI_URL', 'OMEGA_QUALIFIER']
    keys = keys or []
    auth_env = AuthenticationEnv.secure()
    with open(configfile, 'w') as fconfig:
        configs = auth_env.get_userconfig_from_api(api_url=api_url,
                                                   userid=userid,
                                                   apikey=apikey,
                                                   requested_userid=requested_userid,
                                                   qualifier=qualifier,
                                                   view=view)
        config = configs['objects'][0]['data']
        config['OMEGA_RESTAPI_URL'] = api_url
        config['OMEGA_QUALIFIER'] = qualifier or 'default'
        config['OMEGA_USERID'] = userid
        config['OMEGA_APIKEY'] = apikey
        config = {
            k: v for k, v in config.items() if k in (required_keys + keys)
        }
        yaml.safe_dump(config, fconfig, default_flow_style=False)
        print("Config is in {configfile}".format(**locals()))



import base64
from unittest import TestCase

import json
from nbformat import NotebookNode
from tornado.web import HTTPError

from omegaml import Omega
from omegaml.notebook.omegacontentsmgr import OmegaStoreContentsManager
from omegaml.tests.util import OmegaTestMixin


class OmegaContentsManagerTests(OmegaTestMixin, TestCase):
    def setUp(self):
        self.om = Omega()
        self.mgr = OmegaStoreContentsManager(omega=self.om)
        self.clean()

    def _create_notebook(self, name, content=None):
        code = content or """
        print('hello world')
        """.strip()
        self.om.jobs.create(code, name)

    def test_get_top_level(self):
        om = self.om
        # check director is empty to start with
        model = self.mgr.get('/')
        self.assertEqual(len(model['content']), 0)
        # add a new model, check it is represented
        self._create_notebook('foo')
        model = self.mgr.get('/')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'foo.ipynb')
        self.assertEqual(model['content'][0]['type'], 'notebook')

    def test_subdirectory(self):
        om = self.om
        # create a notebook in a sub directory
        self._create_notebook('sub/foo')
        model = self.mgr.get('/')
        # check we can see the directory
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['type'], 'directory')
        self.assertEqual(model['content'][0]['name'], 'sub')
        # check we can see the notebook
        model = self.mgr.get('/sub', type='directory')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'foo.ipynb')
        # create a notebook in a sub sub directory
        self._create_notebook('sub1/sub2/foo')
        model = self.mgr.get('/sub1', type='directory')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['type'], 'directory')
        self.assertEqual(model['content'][0]['name'], 'sub2')
        model = self.mgr.get('/sub1/sub2', type='directory')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'foo.ipynb')

    def test_save_notebook(self):
        # create a dummy notebook just so we have a valid model
        self._create_notebook('foo')
        nbmodel = self.mgr.get('foo.ipynb', type='notebook')
        # copy foo to bar
        model = self.mgr._base_model('sub', kind='directory')
        self.mgr.save(model, 'sub')
        model = self.mgr._notebook_model('foo.ipynb')
        self.mgr.save(model, 'sub/bar.ipynb')
        # check it got created within a directory
        model = self.mgr.get('/sub', type='directory')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'bar.ipynb')
        # check it can be read directly
        model = self.mgr.get('/sub/bar.ipynb', type='notebook')
        self.assertEqual(model['name'], 'bar.ipynb')
        self.assertEqual(model['content']['cells'][0]['source'], "print('hello world')")

    def test_blankfile_exists_always_false(self):
        # this protects against cannot use POST, use PUT error
        # see https://stackoverflow.com/a/59775912/890242
        # create a first notebook
        result = self.mgr.file_exists('')
        self.assertEqual(result, False)
        self._create_notebook('foo')
        result = self.mgr.file_exists('')
        self.assertEqual(result, False)
        result = self.mgr.file_exists('foo2')
        self.assertEqual(result, False)
        self._create_notebook('sub/foo')
        result = self.mgr.file_exists('sub')
        self.assertEqual(result, False)

    def test_get_notebook_as_file(self):
        # create a dummy notebook just so we have a valid model
        self._create_notebook('foo')
        # check it can be read directly as a file
        model = self.mgr.get('foo.ipynb', type='file')
        nb_content = json.loads(model['content'])
        self.assertEqual(model['name'], 'foo.ipynb')
        self.assertEqual(nb_content['cells'][0]['source'], ["print('hello world')"])

    def test_save_notebook_as_file(self):
        # create a dummy notebook just so we have a valid model
        self._create_notebook('foo')
        # check it can be read directly as a file
        model = self.mgr.get('foo.ipynb', type='file')
        model['name'] = 'bar.ipynb'
        self.mgr.save(model, 'bar.ipynb')
        self.assertIn('bar.ipynb', self.om.jobs.list())
        nb = self.om.jobs.get('bar.ipynb')
        self.assertIsInstance(nb, NotebookNode)

    def test_save_notebook_as_file_base64(self):
        # create a dummy notebook just so we have a valid model
        self._create_notebook('foo')
        # check it can be read directly as a file
        model = self.mgr.get('foo.ipynb', type='file', format='base64')
        model['name'] = 'bar.ipynb'
        model['format'] = 'base64'
        model['content'] = base64.encodebytes(json.dumps(model['content']).encode('utf8'))
        self.mgr.save(model, 'bar.ipynb')
        self.assertIn('bar.ipynb', self.om.jobs.list())
        nb = self.om.jobs.get('bar.ipynb')
        self.assertIsInstance(nb, NotebookNode)

    def test_overwrite_existing_notebook(self):
        # create a dummy notebook just so we have a valid model
        self._create_notebook('foo')
        nbmodel = self.mgr.get('foo.ipynb', type='notebook')
        # copy foo to bar
        model = self.mgr._base_model('sub', kind='directory')
        self.mgr.save(model, 'sub')
        model = self.mgr._notebook_model('foo.ipynb')
        self.mgr.save(model, 'sub/foo.ipynb')
        # save again
        model['content']['cells'][0]['source'] = "print('hello world at large!')"
        self.mgr.save(model, 'sub/foo.ipynb')
        # check it got created within a directory
        model = self.mgr.get('/sub', type='directory')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'foo.ipynb')
        # check it can be read directly
        model = self.mgr.get('/sub/foo.ipynb', type='notebook')
        self.assertEqual(model['name'], 'foo.ipynb')
        self.assertEqual(model['content']['cells'][0]['source'], "print('hello world at large!')")

    def test_create_directory(self):
        # test creating directory on top level
        model = self.mgr.get('/')
        self.assertEqual(len(model['content']), 0)
        model = self.mgr._base_model('sub', kind='directory')
        self.mgr.save(model, 'sub')
        model = self.mgr.get('/')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'sub')
        self.assertEqual(model['content'][0]['type'], 'directory')
        # test creating directory on sub level
        model = self.mgr._base_model('sub/sub1', kind='directory')
        self.mgr.save(model, 'sub/sub1')
        model = self.mgr.get('/')
        self.assertEqual(len(model['content']), 1)
        model = self.mgr.get('/sub', type='directory')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'sub1')
        self.assertEqual(model['content'][0]['type'], 'directory')

    def test_get_directory_contents(self):
        # test creating directory on top level
        model = self.mgr._base_model('sub', kind='directory')
        self.mgr.save(model, 'sub')
        model = self.mgr.get('/')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'sub')
        self.assertEqual(model['content'][0]['type'], 'directory')
        # test getting empty directory
        model = self.mgr.get('/sub')
        self.assertEqual('sub', model['name'])
        self.assertEqual('directory', model['type'])
        self.assertEqual([], model['content'])

    def test_create_directory_escaped(self):
        # test creating directory on top level
        model = self.mgr._base_model('Untitled Folder', kind='directory')
        self.mgr.save(model, 'Untitled Folder')
        model = self.mgr.get('/')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'Untitled Folder')
        self.assertEqual(model['content'][0]['type'], 'directory')
        # test getting it with escaped string
        self.assertTrue(self.mgr.exists('Untitled Folder'))
        self.assertTrue(self.mgr.exists('Untitled%20Folder'))

    def test_create_notebook_in_empty_directory(self):
        # test creating directory on top level
        model = self.mgr._base_model('sub', kind='directory')
        self.mgr.save(model, 'sub')
        model = self.mgr.get('/')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'sub')
        self.assertEqual(model['content'][0]['type'], 'directory')
        result = self.mgr.exists('sub')
        self.assertTrue(result)
        # test creating a notebook in this new emtpty directory
        self._create_notebook('sub/foo.ipynb')
        model = self.mgr.get('sub/foo.ipynb', type='notebook')
        self.assertEqual(model['name'], 'foo.ipynb')
        self.assertEqual(model['content']['cells'][0]['source'], "print('hello world')")

    def test_create_multiple_notebook_in_empty_directory(self):
        # test creating directory on top level
        model = self.mgr._base_model('sub', kind='directory')
        self.mgr.save(model, 'sub')
        model = self.mgr.get('/')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'sub')
        self.assertEqual(model['content'][0]['type'], 'directory')
        result = self.mgr.exists('sub')
        self.assertTrue(result)
        # test creating a notebook in this new emtpty directory
        self._create_notebook('sub/foo.ipynb')
        self._create_notebook('sub/foo2.ipynb')
        self._create_notebook('sub/foo3.ipynb')
        model = self.mgr.get('/')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'sub')
        self.assertEqual(model['content'][0]['type'], 'directory')

    def test_exists(self):
        om = self.om
        # check directory is empty to start with
        result = self.mgr.exists('')
        self.assertTrue(result)
        result = self.mgr.dir_exists('')
        self.assertTrue(result)
        result = self.mgr.file_exists('')
        self.assertFalse(result)
        result = self.mgr.file_exists('test.ipynb')
        self.assertFalse(result)
        # check directory is empty to start with
        result = self.mgr.exists('/Untitled1')
        self.assertFalse(result)
        result = self.mgr.exists('/sub/sub1')
        self.assertFalse(result)
        # test creating directory on sub level
        model = self.mgr._base_model('sub/sub1', kind='directory')
        self.mgr.save(model, 'sub/sub1')
        result = self.mgr.exists('sub/sub1')
        self.assertTrue(result)
        # test dropping the directory actually works
        self.mgr.delete('sub/sub1')
        result = self.mgr.exists('sub/sub1')
        self.assertFalse(result)
        # test creating a file works
        model = self.mgr._base_model('Notebook.ipynb', kind='notebook')
        result = self.mgr.exists('Notebook.ipynb')

    def test_rename_notebook(self):
        # create a dummy notebook just so we have a valid model
        self._create_notebook('foo')
        nbmodel = self.mgr.get('foo.ipynb', type='notebook')
        self.mgr.rename('foo.ipynb', 'bar.ipynb')

    def test_rename_directory(self):
        model = self.mgr._base_model('sub', kind='directory')
        self.mgr.save(model, 'sub')
        self.mgr.rename_file('sub', 'new_sub')
        with self.assertRaises(HTTPError):
            model = self.mgr.get('sub', type='directory')
        model = self.mgr.get('new_sub', type='directory')
        self.assertEqual(len(model['content']), 0)

    def test_save_file_base64(self):
        model = self.mgr._base_model('textfile.txt', kind='file')
        model['content'] = base64.encodebytes('hello world'.encode('utf8')).decode('ascii')
        model['format'] = 'base64'
        self.mgr.save(model, 'textfile.txt')
        model = self.mgr.get('textfile.txt', type='file')
        self.assertEqual(model['content'], 'hello world')

    def test_save_file_text(self):
        model = self.mgr._base_model('textfile.txt', kind='file')
        model['content'] = 'hello world'
        model['format'] = 'text'
        self.mgr.save(model, 'textfile.txt')
        model = self.mgr.get('textfile.txt', type='file')
        self.assertEqual(model['content'], 'hello world')

    def test_weird_filenames(self):
        for fn in ('Untitled (1).ipynb', 'Some+strange+name.+++'):
            self._create_notebook(fn)
            expected_fn = fn + ('.ipynb' if not fn.endswith('.ipynb') else '')
            self.assertIn(expected_fn, self.om.jobs.list())
            model = self.mgr.get('/')
            contents = [e['name'] for e in model['content']]
            self.assertIn(expected_fn, contents)

    def test_save_file_1MB_text(self):
        model = self.mgr._base_model('textfile.txt', kind='file')
        very_large = 'abc' * 1024 * 1024
        model['content'] = very_large
        model['format'] = 'text'
        self.mgr.save(model, 'textfile.txt')
        model = self.mgr.get('textfile.txt', type='file')
        self.assertEqual(model['content'], very_large)

    def test_save_notebook_1MB(self):
        very_large = 'abc' * 1024 * 1024
        self._create_notebook('foo', content=very_large)
        nbmodel = self.mgr.get('foo.ipynb', type='notebook')



import unittest
from datetime import datetime

import pandas as pd
from pandas._testing import assert_frame_equal

from omegaml import Omega
from omegaml.mixins.store.datarevision import DataRevisionMixin
from omegaml.tests.util import OmegaTestMixin


class DataRevisionMixinTests(OmegaTestMixin, unittest.TestCase):
    def setUp(self):
        self.om = Omega()
        self.clean()
        self.om.datasets.register_mixin(DataRevisionMixin)

    def test_revision_store_retrieve(self):
        """ check storing revisions works as expected"""
        om = self.om
        # storing dataset by numeric index, 0-9
        df_a = pd.DataFrame({
            'x': range(0, 10)
        })
        meta = om.datasets.put(df_a, 'revtest', append=False, revisions=True)
        self.assertIn('revisions', meta.kind_meta)
        self.assertEqual(meta.kind_meta['revisions']['seq'], 0)
        # update idx data 0-5 with new data
        df_b = pd.DataFrame({
            'x': range(5, 10)
        })
        meta = om.datasets.put(df_b, 'revtest')
        self.assertEqual(meta.kind_meta['revisions']['seq'], 1)
        # check we retrieve updated frame as the latest revision
        dfx = om.datasets.get('revtest')
        assert_frame_equal(dfx.iloc[0:5], df_b.iloc[0:5])
        assert_frame_equal(dfx.iloc[5:], df_a.iloc[5:])
        # check we can get a specific version
        dfx = om.datasets.get('revtest', revision=0)
        assert_frame_equal(dfx, df_a)
        # check the data is clean from revision internals
        self.assertNotIn('_om#revision', dfx.columns)
        self.assertNotIn('_om#revision', df_a)
        self.assertNotIn('_delete_', dfx.columns)
        self.assertNotIn('_delete_', df_a)
        # check we can get a specific changeset
        dfx = om.datasets.get('revtest', changeset=1)
        assert_frame_equal(dfx, df_b)
        # check we can append new data and get back revisions with and without this
        df_c = pd.DataFrame({
            'x': range(11, 15)
        }, index=range(11, 15))
        meta = om.datasets.put(df_c, 'revtest')
        dfx = om.datasets.get('revtest', revision=-1)
        assert_frame_equal(dfx.iloc[5:10], df_a.iloc[5:10])
        assert_frame_equal(dfx.iloc[0:5], df_b.iloc[0:5])
        assert_frame_equal(dfx.iloc[10:], df_c.iloc[0:])
        # check we can update previously appended data
        df_d = pd.DataFrame({
            'x': range(16, 18)
        }, index=range(12, 14))
        meta = om.datasets.put(df_d, 'revtest')
        dfx = om.datasets.get('revtest', revision=-1)
        assert_frame_equal(dfx.iloc[5:10], df_a.iloc[5:10])
        assert_frame_equal(dfx.iloc[0:5], df_b.iloc[0:5])
        assert_frame_equal(dfx.iloc[10:11], df_c.iloc[0:1])
        assert_frame_equal(dfx.iloc[11:13], df_d.iloc[0:])
        assert_frame_equal(dfx.iloc[13:], df_c.iloc[3:])

    def test_revision_store_retrieve_multiple(self):
        om = self.om
        # storing dataset by numeric index, 0-9
        df_a = pd.DataFrame({
            'x': range(0, 10)
        }, index=range(0, 10))
        meta = om.datasets.put(df_a, 'revtest', append=False, revisions=True)
        # update with new data
        df_b = pd.DataFrame({
            'x': range(3, 6)
        }, index=range(3, 6))
        meta = om.datasets.put(df_b, 'revtest')
        # update with new data
        df_c = pd.DataFrame({
            'x': range(6, 10)
        }, index=range(6, 10))
        meta = om.datasets.put(df_c, 'revtest')
        # get back original revision
        dfx = om.datasets.get('revtest', revision=0)
        assert_frame_equal(dfx, df_a)
        # get back second revision
        dfx = om.datasets.get('revtest', revision=1)
        assert_frame_equal(dfx.iloc[0:3], df_a.iloc[0:3])
        assert_frame_equal(dfx.iloc[3:6], df_b.iloc[0:])
        assert_frame_equal(dfx.iloc[6:], df_a.iloc[6:])
        # get back most recent
        dfx = om.datasets.get('revtest', revision=2)
        assert_frame_equal(dfx.iloc[0:3], df_a.iloc[0:3])
        assert_frame_equal(dfx.iloc[3:6], df_b.iloc[0:])
        assert_frame_equal(dfx.iloc[6:], df_c.iloc[0:])

    def test_revisions_bydate(self):
        om = self.om
        # storing dataset by numeric index, 0-9
        df_a = pd.DataFrame({
            'x': range(0, 10)
        })
        dt_a = datetime.utcnow()
        meta = om.datasets.put(df_a, 'revtest', append=False, revisions=True, revision_dt=dt_a)
        # update idx data 0-5 with new data
        df_b = pd.DataFrame({
            'x': range(5, 10)
        })
        dt_b = datetime.utcnow()
        meta = om.datasets.put(df_b, 'revtest', revision_dt=dt_b)
        # get back original revision by date
        dfx = om.datasets.get('revtest', revision=dt_a)
        assert_frame_equal(dfx, df_a)
        # get back most recent by date
        dfx = om.datasets.get('revtest', revision=dt_b)
        assert_frame_equal(dfx.iloc[0:5], df_b.iloc[0:5])
        assert_frame_equal(dfx.iloc[5:], df_a.iloc[5:])
        # retrieve by string-specified date
        dfx = om.datasets.get('revtest', revision=dt_a.isoformat())
        assert_frame_equal(dfx, df_a)

    def test_revisions_bytag(self):
        om = self.om
        # storing dataset by numeric index, 0-9
        df_a = pd.DataFrame({
            'x': range(0, 10)
        })
        meta = om.datasets.put(df_a, 'revtest', append=False, revisions=True,
                               tag='rev_a')
        # get back original revision by atag
        dfx = om.datasets.get('revtest', revision='rev_a')
        assert_frame_equal(dfx, df_a)

    def test_revisions_bynegativeindex(self):
        om = self.om
        # storing dataset by numeric index, 0-9
        df_a = pd.DataFrame({
            'x': range(0, 10)
        }, index=range(0, 10))
        meta = om.datasets.put(df_a, 'revtest', append=False, revisions=True)
        # update with new data
        df_b = pd.DataFrame({
            'x': range(3, 6)
        }, index=range(3, 6))
        meta = om.datasets.put(df_b, 'revtest')
        # update with new data
        df_c = pd.DataFrame({
            'x': range(6, 10)
        }, index=range(6, 10))
        meta = om.datasets.put(df_c, 'revtest')
        # get back original revision
        # -- same as df_a
        dfx = om.datasets.get('revtest', revision=-3)
        assert_frame_equal(dfx, df_a)
        # get back second revision
        # -- parts of df_a and df_b
        dfx = om.datasets.get('revtest', revision=-2)
        assert_frame_equal(dfx.iloc[0:3], df_a.iloc[0:3])
        assert_frame_equal(dfx.iloc[3:6], df_b.iloc[0:])
        assert_frame_equal(dfx.iloc[6:], df_a.iloc[6:])
        # get back most recent
        # -- parts of df_a, df_b, df_c
        dfx = om.datasets.get('revtest', revision=-1)
        assert_frame_equal(dfx.iloc[0:3], df_a.iloc[0:3])
        assert_frame_equal(dfx.iloc[3:6], df_b.iloc[0:])
        assert_frame_equal(dfx.iloc[6:], df_c.iloc[0:])

    def test_revisions_list(self):
        om = self.om
        # storing dataset by numeric index, 0-9
        df_a = pd.DataFrame({
            'x': range(0, 10)
        })
        meta = om.datasets.put(df_a, 'revtest', append=False, revisions=True,
                               tag='rev_a')
        df_b = pd.DataFrame({
            'x': range(0, 2)
        })
        om.datasets.put(df_b, 'revtest')
        # check revisions list
        revs = om.datasets.revisions('revtest')
        self.assertEqual(len(revs), 2)
        self.assertEqual(['dt', 'seq', 'tags', 'delete'], list(revs.columns))

    def test_revisions_delete_byindex(self):
        om = self.om
        # storing dataset by numeric index, 0-9
        df_a = pd.DataFrame({
            'x': range(0, 10)
        })
        meta = om.datasets.put(df_a, 'revtest', append=False, revisions=True)
        df_b = pd.DataFrame({
            'x': range(5, 8),
        }, index=range(5, 8))
        # apply deletions
        df_b['_delete_'] = True
        meta = om.datasets.put(df_b, 'revtest', tag='with-deletions')
        # check deletion happened
        dfx = om.datasets.get('revtest')
        self.assertEqual(len(dfx), len(df_a) - len(df_b))
        assert_frame_equal(dfx.iloc[0:5], df_a.iloc[0:5])
        assert_frame_equal(dfx.iloc[5:], df_a.iloc[8:])
        # check deletions are applied on getting a revision
        df_c = pd.DataFrame({
            'x': range(5, 8),
        }, index=range(5, 8))
        # -- add the deleted rows again
        om.datasets.put(df_c, 'revtest')
        dfx = om.datasets.get('revtest', revision=-1)
        self.assertEqual(len(dfx), len(df_a))
        assert_frame_equal(dfx.sort_index()[['x']], df_a[['x']])
        # get back revisions before re-adding, i.e. revisions applied one by one
        dfx = om.datasets.get('revtest', revision='with-deletions')
        self.assertEqual(len(dfx), len(df_a) - len(df_b))
        assert_frame_equal(dfx.iloc[0:5], df_a.iloc[0:5])
        assert_frame_equal(dfx.iloc[5:], df_a.iloc[8:])

    def test_revisions_delete_byflag(self):
        om = self.om
        # storing dataset by numeric index, 0-9
        df_a = pd.DataFrame({
            'x': range(0, 10)
        })
        meta = om.datasets.put(df_a, 'revtest', append=False, revisions=True)
        df_b = pd.DataFrame({
            'x': range(5, 8),
        }, index=range(5, 8))
        # apply deletions
        meta = om.datasets.put(df_b, 'revtest', delete=True, tag='with-deletions')
        # check deletion happened
        dfx = om.datasets.get('revtest')
        self.assertEqual(len(dfx), len(df_a) - len(df_b))
        assert_frame_equal(dfx.iloc[0:5], df_a.iloc[0:5])
        assert_frame_equal(dfx.iloc[5:], df_a.iloc[8:])
        # check deletions are applied on getting a revision
        df_c = pd.DataFrame({
            'x': range(5, 8),
        }, index=range(5, 8))
        # -- add the deleted rows again
        om.datasets.put(df_c, 'revtest')
        dfx = om.datasets.get('revtest', revision=-1)
        self.assertEqual(len(dfx), len(df_a))
        assert_frame_equal(dfx.sort_index()[['x']], df_a[['x']])
        # get back revisions before re-adding, i.e. revisions applied one by one
        dfx = om.datasets.get('revtest', revision='with-deletions')
        self.assertEqual(len(dfx), len(df_a) - len(df_b))
        assert_frame_equal(dfx.iloc[0:5], df_a.iloc[0:5])
        assert_frame_equal(dfx.iloc[5:], df_a.iloc[8:])

    def test_revisions_trace(self):
        om = self.om
        # storing dataset by numeric index, 0-9
        df_a = pd.DataFrame({
            'x': range(0, 10)
        })
        meta = om.datasets.put(df_a, 'revtest', append=False, revisions=True,
                               tag='rev_a')
        # get back original revision by atag
        dfx = om.datasets.get('revtest', revision='rev_a', trace_revisions=True)
        assert_frame_equal(dfx[['x']], df_a)
        self.assertIn('_om#revision', dfx.columns)
        self.assertIn('_delete_', dfx.columns)

